import unittest
import math

from .route import *


class RouteTest( unittest.TestCase ):

    def assertEqualCoords( self, p, q, places=6 ):
        self.assertAlmostEqual( p[0], q[0], places )
        self.assertAlmostEqual( p[1], q[1], places )

    def testFlattenConversion( self ):
        conv = Convertor()
        geo = (16.00, 49.00)
        planar = conv.geo2planar( geo )
        geo2 = conv.planar2geo( planar )
        self.assertEqualCoords( geo, geo2 )

    def testFindNearest( self ):
        self.assertEqual( Route().findNearest( (16.0, 49.0) ), None )
        self.assertEqualCoords( Route([(17.0, 51.3)]).findNearest( (16.0, 49.0) ), (17.0, 51.3) )
        r = Route( [(10.0, 51.0), (20.0, 52.0), (30.0, 43.2)] )
        self.assertEqualCoords( r.findNearest( (22.0, 50.2) ), (22.029302716939391, 50.214213605948785) )

    def testFindNearestEx( self ):
        conv = Convertor( (0.0, 0.0) )
        r = Route( conv = conv )
        pts = []
        for i in range(10):
            pts.append( (i, 0.0) )
        for i in range(10):
            pts.append( (10.0, i) )
        r.pts = pts
        (pos, dist, index) = r.findNearestEx( conv.planar2geo( (2.5, 3.0) ) )
        self.assertAlmostEqual( 3.0, dist, 5 )
        self.assertEqual( -3, index )
        self.assertEqual( -1, r.findNearestEx( conv.planar2geo( (0.5, 3.0) ) )[2] )
        self.assertEqual( 0, r.findNearestEx( conv.planar2geo( (-0.5, 3.0) ) )[2] )
        self.assertEqual( 10, r.findNearestEx( conv.planar2geo( (20, -20) ) )[2] )

    def testRouteSplit( self ):
        conv = Convertor( (0.0, 0.0) )
        r = Route( conv = conv )
        pts = []
        for i in range(10):
            pts.append( (i, 0.0) )
        for i in range(11):
            pts.append( (10.0, i) )
        r.pts = pts
        first, second = r.routeSplit( conv.planar2geo( (20, -20) ) )
        self.assertEqual( 11, len(first) )
        self.assertEqual( 11, len(second) )
        self.assertEqualCoords( (10.0,0.0), conv.geo2planar( first[-1] ) )
        self.assertEqualCoords( (10.0,0.0), conv.geo2planar( second[0] ) )
        first, second = r.routeSplit( conv.planar2geo( (2.5, -2.0) ) )
        self.assertEqual( 4, len(first) ) # 0 1 2 2.5
        self.assertEqual( 21-2, len(second) ) # 2.5 3 4 ...

    def testPointAtDist( self ):
        conv = Convertor( (0.0, 0.0) )
        r = Route( conv = conv )
        self.assertEqual( None, r.pointAtDist( 2.5 ) )
        pts = []
        for i in range(10):
            pts.append( (i, 0.0) )
        r.pts = pts
        self.assertEqualCoords( (2.5,0.0), conv.geo2planar(r.pointAtDist( 2.5 )) )
        self.assertEqualCoords( (5.0,0.0), conv.geo2planar(r.pointAtDist( 5.0 )) )
        self.assertEqualCoords( (9.0,0.0), conv.geo2planar(r.pointAtDist( 500.0 )) )
        self.assertEqualCoords( (0.0,0.0), conv.geo2planar(r.pointAtDist( 0.0 )) )

    def testTurnAngleAt( self ):
        conv = Convertor( (0.0, 0.0) )
        r = Route( conv = conv )
        pts = []
        for i in range(10):
            pts.append( (i, 0.0) )
        for i in range(11):
            pts.append( (10.0, i) )
        for i in range(10):
            pts.append( (10.0+i, 10.0) )
        r.pts = pts
        self.assertAlmostEqual( 0.0, r.turnAngleAt( conv.planar2geo( (2, -2) ) ), 5 )
        self.assertAlmostEqual( math.pi/2, r.turnAngleAt( conv.planar2geo( (10, 0) ) ), 5 )
        self.assertAlmostEqual( -math.pi/2, r.turnAngleAt( conv.planar2geo( (10, 10) ) ), 5 )
        r.pts.reverse()
        self.assertAlmostEqual( 0.0, r.turnAngleAt( conv.planar2geo( (2, -2) ) ), 5 )
        self.assertAlmostEqual( -math.pi/2, r.turnAngleAt( conv.planar2geo( (10, 0) ) ), 5 )
        self.assertAlmostEqual( math.pi/2, r.turnAngleAt( conv.planar2geo( (10, 10) ) ), 5 )

        # behind EndOfRoute (problem detected during testing Malesice090917
        self.assertAlmostEqual( 0.0, r.turnAngleAt( conv.planar2geo( (-20, 0) ) ), 5 )
        self.assertAlmostEqual( 0.0, r.turnAngleAt( conv.planar2geo( (100, 100) ) ), 5 )
        self.assertFalse(r.isLoop)

    def testIsLoop( self ):
        conv = Convertor( (0.0, 0.0) )
        pts = []
        for i in range(10):
            pts.append( (i, 0.0) )
        for i in range(10):
            pts.append( (10.0, i) )
        for i in range(10):
            pts.append( (10.0-i, 10.0) )
        for i in range(10):
            pts.append( (0.0, 10.0-i) )

        r = Route( [conv.planar2geo(x) for x in pts], conv = conv )
        self.assertTrue( r.isLoop )
        self.assertAlmostEqual( math.pi/2, r.turnAngleAt( conv.planar2geo( (-1, -1) ) ), 5 )

    def testMalesice090917LastRound( self ):
        tomasOkruh = """
# Converted from TomasOkruh.kml
50.084366 14.498719
50.084271 14.498445
50.084502 14.498133
50.084654 14.497916
50.084673 14.497982
50.084676 14.498478
50.084708 14.498862
50.085279 14.498498
50.085586 14.498983
50.085407 14.499124
50.085356 14.499679 
"""
        r = Route( loadLatLonPtsFromStr( tomasOkruh ) )
        self.assertTrue( r.length() < 1000 )
        self.assertTrue( r.findNearestEx( (14.498925, 50.084739999999996) )[2] < 10 )

    def testLine( self ):
        line = Line( (1,2), (3,4) )
        self.assertAlmostEqual( line.signedDistance( (2,3) ), 0.0, 10 )
        self.assertTrue( line.signedDistance( (0, 10) ) > 0 )

    def testLineFinished( self ):
        line = Line( (0,0), (10,0) )
        self.assertTrue( not line.finished( ( 5.0, 0.0 ) ) )
        self.assertTrue( line.finished( ( 11.0, 0.0 ) ) )
        self.assertTrue( line.finished( ( 11.0, 1.0 ) ) )

    def testLineSnap( self ):
        self.assertEqualCoords( Line( (0,0), (10,0) ).snap( (2,3) ), (2,0) )

    def testLineNearest( self ):
        self.assertEqual( Line( (1,2), (3,4) ).nearest( (7,7) ), ( (3,4), 5.0, 1 ) )
        self.assertEqual( Line( (1,2), (3,4) ).nearest( (0,0) ), ( (1,2), math.sqrt(5), 0 ) )
        pos, dist, type = Line( (1,2), (3,4) ).nearest( (2,3) )
        self.assertEqualCoords( pos, (2,3) )
        self.assertAlmostEqual( dist, 0.0 )
        self.assertEqual( type , -1 )


# vim: expandtab sw=4 ts=4



import unittest
from unittest.mock import MagicMock, call

from subt.marsupial import Marsupial
from osgar.lib import quaternion

class MarsupialTest(unittest.TestCase):

    def test_detach(self):
        bus = bus=MagicMock()
        robot = Marsupial(bus=bus, config={'release_at':10})
        robot.on_sim_time_sec(3)
        bus.publish.assert_not_called()
        robot.on_sim_time_sec(14)
        bus.publish.assert_called()
        bus.publish.assert_has_calls([call('detach', [])])
        bus.publish.reset_mock()
        robot.on_sim_time_sec(15)
        bus.publish.assert_not_called()

    def test_origin(self):
        bus = bus=MagicMock()
        robot = Marsupial(bus=bus, config={})
        self.assertIsNone(robot.release_at)
        origin_data = [b'A10W100L', -6.410452, 5.000274, 0.807016, -0.005923, -1.8e-05, 0.999982, -2.4e-05]
        robot.on_origin(origin_data)
        self.assertEqual(robot.release_at, 10)

# vim: expandtab sw=4 ts=4



"""
  Driver "replay" replaying old logfile
"""

from threading import Thread

from osgar.logger import LogReader, lookup_stream_names
from osgar.lib.serialize import deserialize


class ReplayDriver:
    def __init__(self, config, bus):
        self.input_thread = Thread(target=self.run_input, daemon=True)
        self.bus = bus

        self.filename = config['filename']
        self.pins = config['pins']

    def start(self):
        self.input_thread.start()

    def join(self, timeout=None):
        self.input_thread.join(timeout=timeout)

    def run_input(self):
        names = lookup_stream_names(self.filename)
        print(names)
        ids = [i + 1 for i, name in enumerate(names) if name in self.pins]
        print(ids)
        for timestamp, channel_index, data_raw in LogReader(self.filename,
                only_stream_id=ids):
            if not self.bus.is_alive():
                break
            channel = names[channel_index - 1]
            assert channel in self.pins
            data = deserialize(data_raw)
            # TODO reuse timestamp
            self.bus.publish(self.pins[channel], data)
        print('Replay completed!')

    def request_stop(self):
        self.bus.shutdown()

# vim: expandtab sw=4 ts=4



import unittest
from unittest.mock import patch, MagicMock

from osgar.drivers.logserial import LogSerial
from osgar.bus import Bus


class LogSerialTest(unittest.TestCase):

    def test_twoway_communication(self):
        with patch('osgar.drivers.logserial.serial.Serial') as mock:
            instance = mock.return_value

            logger = MagicMock()
            bus = Bus(logger)
            config = {'port':'COM13:', 'speed':4800}
            device = LogSerial(config=config, bus=bus.handle('serial'))
            tester = bus.handle('tester')
            tester.register('raw')
            bus.connect('tester.raw', 'serial.raw')
            tester.publish('raw', b'bin data')
            device.start()
            device.request_stop()
            device.join()
            instance.write.assert_called_once_with(b'bin data')

    def test_timeout_config(self):
        with patch('osgar.drivers.logserial.serial.Serial', autospec=True) as mock:
            instance = mock.return_value
            bus = MagicMock()
            config = {'port':'COM13:', 'speed':4800, 'timeout':2.0}
            device = LogSerial(config=config, bus=bus)
            mock.assert_called_once_with('COM13:', 4800)
            self.assertAlmostEqual(instance.timeout, 2.0)

    def test_config_reset(self):
        with patch('osgar.drivers.logserial.serial.Serial', autospec=True) as mock:
            instance = mock.return_value
            bus = MagicMock()
            config = {'port':'COM10:', 'speed':9600, 'rtscts':True, 'reset':True}
            device = LogSerial(config=config, bus=bus)
            mock.assert_called_once_with('COM10:', 9600, rtscts=True)
            instance.setRTS.assert_called_once_with()
            instance.setDTR.assert_called_once_with(0)

# vim: expandtab sw=4 ts=4



"""
  Communication module for vESC - open source BLDC motor driver
  https://www.vandaelectronics.com/collections/electronic-speed-controls/products/vesc
  https://github.com/vedderb/bldc

  There are several ways how to communicate with the device and here
  we work with CAN bus. The motors uses extended addresing mode (24bits)
  and combines command with device identification into the address.

  At the moment only one type of message is decoded and that is "read buffer".
  The device responds with provided "last_message_id" encoded in the address.
  As we need to communicate with four BLDC motors we abuse this "last_message_id"
  for motor identification, so to query motor #1 we set "last_message_id" to 1 etc.
  The reponse is then collection of several CAN messages with address 0x5nn, where nn
  is the motor number, i.e. 0x501 for motor #1.

  Each 0x5nn message contains offset byte in final buffer and seven data bytes (CAN messages
  are limited to 8 bytes only). 

  The transmission is terminated with 0x7nn message, which carries control chechsum and
  total length of already transmitted data.

  Note that the status structure varies among versions and now we are experimenting with
  version 3.40 only.

  For more info see:
     https://robotika.cz/articles/reviews/vesc
"""

import struct
from datetime import timedelta

from osgar.node import Node
from osgar.bus import BusShutdownException


def draw_diff(arr):
    import matplotlib.pyplot as plt
    t = [a[0] for a in arr]
    values = [a[1:] for a in arr]

    line = plt.plot(t, values, '-o', linewidth=2)

    plt.xlabel('time (s)')
    plt.show()


class MotorDriverVESC(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('can')
        self.debug_arr = []
        self.prev = {}
        self.verbose = False  # TODO move into Node
        self.packet = [[], [], [], []]
        self.tachometer = [None, None, None, None]  # for motors #1, #2, #3 and #4

    def update(self):
        channel = super().update()
        assert channel == 'can', channel
        msg_id, data, flags = self.can  # via update()
        if (msg_id & 0xF00) == 0x500:
            motor_index = (msg_id & 0xFF) - 1  # reindexed motors from #1 to array index 0
            assert data[0] % 7 == 0, data[0]
            self.packet[motor_index].extend(data[1:])
            # the structure for version 3.40 has 59 bytes and looks like:
            # https://github.com/vedderb/bldc/blob/43c3bbaf91f5052a35b75c2ff17b5fe99fad94d1/commands.c#L166-L185
            """
     case COMM_GET_VALUES:
     ind = 0;
 0   send_buffer[ind++] = COMM_GET_VALUES;
 1   buffer_append_float16(send_buffer, mc_interface_temp_fet_filtered(), 1e1, &ind);
 3   buffer_append_float16(send_buffer, mc_interface_temp_motor_filtered(), 1e1, &ind);
 5   buffer_append_float32(send_buffer, mc_interface_read_reset_avg_motor_current(), 1e2, &ind);
 9   buffer_append_float32(send_buffer, mc_interface_read_reset_avg_input_current(), 1e2, &ind);
13   buffer_append_float32(send_buffer, mc_interface_read_reset_avg_id(), 1e2, &ind);
17   buffer_append_float32(send_buffer, mc_interface_read_reset_avg_iq(), 1e2, &ind);
21   buffer_append_float16(send_buffer, mc_interface_get_duty_cycle_now(), 1e3, &ind);
23   buffer_append_float32(send_buffer, mc_interface_get_rpm(), 1e0, &ind);
27   buffer_append_float16(send_buffer, GET_INPUT_VOLTAGE(), 1e1, &ind);
29   buffer_append_float32(send_buffer, mc_interface_get_amp_hours(false), 1e4, &ind);
33   buffer_append_float32(send_buffer, mc_interface_get_amp_hours_charged(false), 1e4, &ind);
37   buffer_append_float32(send_buffer, mc_interface_get_watt_hours(false), 1e4, &ind);
41   buffer_append_float32(send_buffer, mc_interface_get_watt_hours_charged(false), 1e4, &ind);
45   buffer_append_int32(send_buffer, mc_interface_get_tachometer_value(false), &ind);
49   buffer_append_int32(send_buffer, mc_interface_get_tachometer_abs_value(false), &ind);
53   send_buffer[ind++] = mc_interface_get_fault();
54   buffer_append_float32(send_buffer, mc_interface_get_pid_pos_now(), 1e6, &ind);
58   send_buffer[ind++] = app_get_configuration()->controller_id;
59
commands_send_packet(send_buffer, ind);
"""
            # the "tachometer" data are stored as 32bit signed integer starting
            # from 45th byte. Because the data blocks contain 7 bytes we are
            # interested in block with "address pointer" set to 42.
            # Then we do not have to re-contruct all four independent bufferes
            # but only "pick" the interesting sub-packet. Yes, there is a danger
            # that checksum of the whole buffer was not valid!
            if data[0] == 42:  # optimization - pick only this packet part
                b = bytes(data[1:])
                prev = self.tachometer[motor_index]
                self.tachometer[motor_index] = struct.unpack_from('>i', b, 3)[0]
                if prev != self.tachometer[motor_index] and self.verbose:
                    print(self.time, hex(msg_id), self.tachometer[motor_index])
                tmp = [None] * 4
                tmp[motor_index] = self.tachometer[motor_index]
                if self.verbose:
                    self.debug_arr.append([self.time.total_seconds()] + tmp)
        if (msg_id & 0xF00) == 0x700:
            motor_index = (msg_id & 0xFF) - 1  # reindexed motors from #1 to array index 0
            if len(self.packet[motor_index]) == 19:  # version
                print(self.time, self.packet[motor_index])
            self.packet[motor_index] = []
        if self.verbose > 1:
            print(self.time, hex(msg_id), list(data), flags)
        return channel

    def run(self):
        try:
            self.update()  # initialize self.time
            while True:
                # example to query all variables
                # COMM_GET_VALUES=4
                for rx_buffer_last_id in [1, 2, 3, 4]:
                    self.publish('can', [0x800 + rx_buffer_last_id, bytes([rx_buffer_last_id, 0, 4]), 1])

                start_time = self.time
                count = 0
                while self.time - start_time < timedelta(seconds=0.1):
                    self.update()
                    msg_id = self.can[0]
                    if (msg_id & 0xF00) == 0x700:
                        count += 1
                        if count >= 4:
                            break
                if count != 4 and self.verbose:
                    print(self.time, "vESC: incomplete loop")

        except BusShutdownException:
            pass


    def draw(self):
        """
        Debug Draw
        """
        draw_diff(self.debug_arr)

# vim: expandtab sw=4 ts=4



import unittest
from unittest.mock import MagicMock
import datetime

from subt.offseter import Offseter
from osgar.bus import Bus


class OffseterTest(unittest.TestCase):

    def test_usage(self):
        logger = MagicMock()
        bus = Bus(logger)
        logger.write = MagicMock(return_value=datetime.timedelta(microseconds=9721))
        c = Offseter(bus=bus.handle('offseter'), config={})
        tester = bus.handle('tester')
        tester.register('origin', 'pose3d')
        bus.connect('tester.origin', 'offseter.origin')
        bus.connect('tester.pose3d', 'offseter.pose3d')
        bus.connect('offseter.pose3d', 'tester.pose3d')
        c.start()
        tester.publish('origin', [b'A300W600R', -6.4, 5.0, 0.8, 0.0, 0.0, 1.0, 0])
        tester.publish('pose3d', [[0, 0, 0], [0, 0, 0, 1]])  # first pose is "consumed"
        tester.publish('pose3d', [[-2, -1, 0], [0, 0, 0, 1]])
        tester.publish('pose3d', [[-2, -3, 1], [0, 0, 0.7071068, 0.7071068]])  # 90 deg
        c.request_stop()
        c.join()
        self.assertIsNotNone(c.init_quat)
        xyz, ori = tester.listen()[2]
        self.assertEqual(xyz, [-6.4 + 2, 5.0 + 1, 0.8])
        self.assertEqual(ori, [0.0, 0.0, 1.0, 0])

        xyz, ori = tester.listen()[2]
        self.assertEqual(xyz, [-6.4 + 2, 5.0 + 3, 0.8 + 1])
        self.assertEqual(ori, [0, 0, 0.7071068, -0.7071068])  # 270 deg

# vim: expandtab sw=4 ts=4




"""
  OSGAR breadcrumbs dispenser for Virtual
"""
from ast import literal_eval

from osgar.node import Node
from subt.trace import distance3D


class Breadcrumbs(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register("deploy", "location")

        self.robot_name = config.get('robot_name')
        self.start_time = None  # unknown
        self.next_deploy_time = None
        self.step_time = config.get('step_sec')

        self.radius = config.get('radius')
        self.locations = [[0, 0, 0]]  # all locations + fake for the base
        self.num_avail = config.get('num', 6)  # Freyja, config 2

    def should_deploy(self, xyz):
        if self.radius is None:
            return False
        for loc in self.locations:
            if distance3D(xyz, loc) < self.radius:
                return False
        return True

    def deploy(self):
        if self.num_avail > 0:
            self.publish('deploy', [])  # ROS std_msg/Empty expects empty list as input
            self.num_avail -= 1
            return True
        return False

    def on_sim_time_sec(self, data):
        if self.start_time is None:
            self.start_time = data
            if self.step_time is not None:
                self.next_deploy_time = self.start_time + self.step_time
        if self.next_deploy_time is not None and data > self.next_deploy_time:
            self.next_deploy_time += self.step_time
            self.deploy()

    def on_pose3d(self, data):
        xyz, quat = data
        if self.should_deploy(xyz) and self.deploy():
            self.locations.append(xyz)
            self.publish('location', xyz)

    def on_external(self, data):
        # external location of new breadcrumb
        if data not in self.locations:
            self.locations.append(data)

    def update(self):
        channel = super().update()
        handler = getattr(self, "on_" + channel, None)
        if handler is not None:
            handler(getattr(self, channel))

# vim: expandtab sw=4 ts=4



import subprocess

from osgar.node import Node


def wifi_scan():
    with subprocess.Popen('sudo iwlist wlan0 scanning', shell=True,
            stdout=subprocess.PIPE) as cmd:
        wifiList = []
        for line in cmd.stdout:
            if 'Quality' in str(line):
                signalStrength = int(str(line).split("=")[2].split('dBm')[0])
            elif 'ESSID' in str(line):
                essid = str(line).split('"')[1]
                wifiList.append([essid,signalStrength])
        return wifiList


class WifiSignal(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('wifiscan')
        self.sleep_time = config.get('sleep', 1.0)
        # TODO ssid
        # search SSID

    def run(self):
        while self.is_bus_alive():
            wifi_list = wifi_scan()
            now = self.publish("wifiscan", wifi_list)
            self.sleep(self.sleep_time)


if __name__ == "__main__":
    import time
    while True:
        result = wifi_scan()
        print(result)
        time.sleep(1)

# vim: expandtab sw=4 ts=4




"""
  GPS Navigation for RoboOrienteering 2019
"""
# based on Spide3 "ro2018.py"

import math
from datetime import timedelta

from osgar.lib.mathex import normalizeAnglePIPI
from osgar.node import Node
from osgar.drivers.gps import INVALID_COORDINATES


def geo_length(pos1, pos2):
    "return distance on sphere for two integer positions in milliseconds"
    x_scale = math.cos(math.radians(pos1[0]/3600000))
    scale = 40000000/(360*3600000)
    return math.hypot((pos2[0] - pos1[0])*x_scale, pos2[1] - pos1[1]) * scale


def geo_angle(pos1, pos2):
    if geo_length(pos1, pos2) < 1.0:
        return None
    x_scale = math.cos(math.radians(pos1[0]/3600000))
    return math.atan2(pos2[1] - pos1[1], (pos2[0] - pos1[0])*x_scale)


def latlon2xy(lat, lon):
    return int(round(lon*3600000)), int(round(lat*3600000))


class EmergencyStopException(Exception):
    pass


class EmergencyStopMonitor:
    def __init__(self, robot):
        self.robot = robot

    def update(self, robot):
        if (robot.status & RoboOrienteering2018.EMERGENCY_STOP) == 0:
            raise EmergencyStopException()

    # context manager functions
    def __enter__(self):
        self.callback = self.robot.register(self.update)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.robot.unregister(self.callback)


class GPSNavigation(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('desired_speed')
        self.maxspeed = config.get('maxspeed', 0.5)
        self.goals = [latlon2xy(lat, lon) for lat, lon in config['waypoints']]
        self.last_position = None  # (lon, lat) in milliseconds
        self.last_imu_yaw = None
        self.cmd = [0, 0]
        self.last_position_angle = None  # for angle computation from dGPS
        self.monitors = []

    def send_speed_cmd(self, speed, angular_speed):
        return self.publish('desired_speed', [round(speed*1000), round(math.degrees(angular_speed)*100)])

    def update(self):
        channel = super().update()  # define self.time
        if channel == 'position':
            self.last_position = self.position
        elif channel == 'rot':  # should be rather 'rotation'
            yaw, pitch, roll = self.rot
            self.last_imu_yaw = math.radians(yaw/100.0)

    def register(self, callback):
        self.monitors.append(callback)
        return callback

    def unregister(self, callback):
        assert callback in self.monitors
        self.monitors.remove(callback)

    def wait(self, dt):
        if self.time is None:
            self.update()
        start_time = self.time
        while self.time - start_time < dt:
            self.update()

    def run(self):
        print("Waiting for valid GPS position...")
        while self.last_position is None or self.last_position == INVALID_COORDINATES:
            self.update()
        print(self.last_position)

        print("Wait for valid IMU...")
        while self.last_imu_yaw is None:
            self.update()
        print(self.last_imu_yaw)

        print("Ready", self.goals)
        try:
            with EmergencyStopMonitor(self):
                for goal in self.goals:
                    print("Goal at %.2fm" % geo_length(self.last_position, goal))
                    angle = geo_angle(self.last_position, goal)
                    if angle is not None:
                        print("Heading %.1fdeg, imu %.1fdeg" % (math.degrees(angle), math.degrees(self.last_imu_yaw)))
                    else:
                        print("Heading None, imu %.1fdeg" % math.degrees(self.last_imu_yaw))
                    self.navigate_to_goal(goal, timedelta(seconds=600))
        except EmergencyStopException:
            print("EMERGENCY STOP (wait 3s)")
            self.send_speed_cmd(0, 0)
            start_time = self.time
            while self.time - start_time < timedelta(seconds=3):
                self.update()

    def navigate_to_goal(self, goal, timeout):
        start_time = self.time
        self.last_position_angle = self.last_position
        gps_angle = None
        while geo_length(self.last_position, goal) > 1.0 and self.time - start_time < timeout:
            desired_heading = normalizeAnglePIPI(geo_angle(self.last_position, goal))
            step = geo_length(self.last_position, self.last_position_angle)
            if step > 1.0:
                gps_angle = normalizeAnglePIPI(geo_angle(self.last_position_angle, self.last_position))
                print('step', step, math.degrees(gps_angle))
                self.last_position_angle = self.last_position
            desired_wheel_heading = normalizeAnglePIPI(desired_heading - self.last_imu_yaw)
            self.send_speed_cmd(self.maxspeed, desired_wheel_heading)

            prev_time = self.time
            self.update()

            if int(prev_time.total_seconds()) != int(self.time.total_seconds()):
                print(self.time, geo_length(self.last_position, goal), math.degrees(self.last_imu_yaw))

        print("STOP (3s)")
        self.send_speed_cmd(0, 0)
        start_time = self.time
        while self.time - start_time < timedelta(seconds=3):
            self.update()


if __name__ == "__main__":
    from osgar.launcher import launch

    launch(app=GPSNavigation, description='GPS Navigation', prefix='gpsnav-')

# vim: expandtab sw=4 ts=4



"""
  Report and filter shared detected artifacts

The artifacts are reported to "DARPA base station" as [type, position]. The "DARPA base station" answers with
message, where only "score_changed" has value 1 if it was correctly localized artifact within 5 meters radius (3D).

The implemented solution is using shared array of items [type, position, source, scored]. Type and 3D position
is the same as for DARPA report. "source" is identification of robot which detected artifact (for example "A1300L").
"scored" has states: True/False/None. This corresponds to base station answer "score_changed"=1/"score_changed"=0/
"unknown=no answer".

If ArtifactReporter receives message with scored=None then first verifies if identical type and position is already
in the pool. If yes, then received message is ignored. There still can be another report already confirmed as
scored=True so it goes through identical types and if position is within RADIUS then it is rejected (it would need
4th type - detected but ignored). Note, that in the rules "SubT_Challenge_Cave_Rules.pdf" is nothing about proximity
of placed artifacts, so there can be two backpacks side by side. RADIUS is only "probabilistic", that it is rather
duplicity than two artifacts placed side by side.

Artifacts received with scored=True/False replace only scored=None in the pool. Not matching values should assert.

There is a race condition between reporting artifact and spreading this info to other robots. It has to be done
immediately otherwise other robots may use this time to report independently new found artifact. The only solution
would be teambase, but then it would be the single point of failure.
"""
from osgar.node import Node
from subt.trace import Trace, distance3D


RADIUS = 4.0  # there are probably not two artifacts within sphere of this radius


class ArtifactReporter(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register("artf_cmd", "artf_all")
        self.repeat_report_sec = config.get('repeat_report_sec')
        self.artf_xyz = []
        self.artf_xyz_accumulated = []  # items are [type, position, source, flag]
                                        # position = [x, y, z], flag(scored) = True/False/None

    def publish_artf(self, artf_xyz):
        count = 0
        for artf_type, pos, src, scored in artf_xyz:
            if scored is None:
                if count == 0:
                    print(self.time, "DETECTED:")
                count += 1
                ix, iy, iz = pos
                print(" ", artf_type, ix/1000.0, iy/1000.0, iz/1000.0)
                s = '%s %.2f %.2f %.2f\n' % (artf_type, ix/1000.0, iy/1000.0, iz/1000.0)
                self.publish('artf_cmd', bytes('artf ' + s, encoding='ascii'))
        if count > 0:
            print('report completed')

    def nearest_scored_artifact(self, artf_type, position):
        # TODO remove 1000x scaling to millimeters - source of headache
        dist = [distance3D(position, artf[1]) for artf in self.artf_xyz_accumulated
                if artf[0] == artf_type and artf[-1] is True]
        if len(dist) == 0:
            return None  # no nearest for given filter
        return min(dist)/1000.0  # due to mm scaling

    def on_sim_time_sec(self, data):
        if self.repeat_report_sec is None or self.sim_time_sec % self.repeat_report_sec != 0:
            return
        if len(self.artf_xyz) == 0:
            return
        # publish all artifacts only on regular update (to avoid LoRA cross-talk)
        self.publish_artf(self.artf_xyz_accumulated)
        self.publish('artf_all', self.artf_xyz_accumulated)

    def on_artf_xyz(self, data):
        for item in data:
            atype, pos, src, scored = item
            if [atype, pos] not in [arr[:2] for arr in self.artf_xyz_accumulated]:
                # new entry with unique (type, position)
                nearest = self.nearest_scored_artifact(atype, pos)
                if nearest is None or nearest > RADIUS:
                    self.artf_xyz_accumulated.append(item)
            elif item not in self.artf_xyz_accumulated:
                # existing entry but with different flag (or source)
                i = [arr[:2] for arr in self.artf_xyz_accumulated].index(item[:2])
                if scored is True or scored is False:
                    assert self.artf_xyz_accumulated[i][-1] is None, self.artf_xyz_accumulated[i]
                    self.artf_xyz_accumulated[i] = item
                else:
                    pass  # item.score is None, but we already know the answer -> ignore
        self.publish_artf(self.artf_xyz_accumulated)

    def on_base_station(self, data):
        p = data['artifact_position']
        dist = [distance3D(p, [x/1000.0 for x in artf[1]]) for artf in self.artf_xyz_accumulated]
        if len(dist) > 0 and min(dist) < 0.1:
            # i.e. it is our report and we have it in the list ... I would almost assert it
            min_i = dist.index(min(dist))
            was_unknown = self.artf_xyz_accumulated[min_i][-1] is None
            self.artf_xyz_accumulated[min_i][-1] = (data['score_change'] > 0)
            # TODO? check type - we decided to ignore it for Cave Circuit
            if was_unknown:
                self.publish('artf_all', self.artf_xyz_accumulated)  # broadcast new update

    def update(self):
        channel = super().update()  # define self.time
        handler = getattr(self, "on_" + channel, None)
        if handler is not None:
            handler(getattr(self, channel))
        return channel


# vim: expandtab sw=4 ts=4



"""
  serialization of messages
"""
from io import BytesIO

import msgpack
import zlib

try:
    import numpy
except:
    numpy = False

MSGPACK_ZLIB_CODE = 42
MSGPACK_NUMPY_ZLIB_CODE = 43 # for old logs only, no longer used
MSGPACK_NUMPY_CODE = 44


def serialize(data):
    if numpy and isinstance(data, (numpy.ndarray, numpy.generic)):
        with BytesIO() as b:
            numpy.save(b, data, allow_pickle=False)
            data = b.getvalue()
        data = msgpack.ExtType(MSGPACK_NUMPY_CODE, data)

    return msgpack.packb(data, use_bin_type=True)


def compress(data):
    zlib_data = zlib.compress(data)
    data = msgpack.ExtType(MSGPACK_ZLIB_CODE, zlib_data)
    return msgpack.packb(data, use_bin_type=True)


def deserialize(bytes_data):
    data = msgpack.unpackb(bytes_data, raw=False)
    while isinstance(data, msgpack.ExtType):
        if data.code == MSGPACK_ZLIB_CODE:
            data = zlib.decompress(data.data)
            data = msgpack.unpackb(data, raw=False)
        elif data.code == MSGPACK_NUMPY_ZLIB_CODE: # for old logs only, no longer used
            if not numpy:
                raise TypeError("TypeError: can not deserialize 'numpy.ndarray' object")
            data = zlib.decompress(data.data)
            data = numpy.load(BytesIO(data), allow_pickle=False)
        elif data.code == MSGPACK_NUMPY_CODE:
            if not numpy:
                raise TypeError("TypeError: can not deserialize 'numpy.ndarray' object")
            data = numpy.load(BytesIO(data.data), allow_pickle=False)
    return data

# vim: expandtab sw=4 ts=4



"""
  Dual robots - typically a UGV taking care of UAV
  https://en.wikipedia.org/wiki/Marsupial

  This marsupial node is necessary for the "child", i.e. drone, which can call "detach" itself.
  There is no need to communicate this with the parent (UGV).
"""
# For more details see https://github.com/osrf/subt/issues/697

from osgar.node import Node
from subt.name_decoder import parse_robot_name


class Marsupial(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register("detach")

        self.start_time = None  # unknown
        self.release_at = config.get('release_at')  # not defined yet
        self.drone_available = True

    def detach(self):
        self.publish('detach', [])  # ROS std_msg/Empty expects empty list as input
        self.drone_available = False

    def on_sim_time_sec(self, data):
        if self.start_time is None:
            self.start_time = data

        if self.release_at is None:
            return

        if self.start_time + self.release_at <= data and self.drone_available:
            self.detach()

    def on_origin(self, data):
        if self.release_at is None:
            robot_name = data[0].decode('ascii')
            print(robot_name)
            cmd_list = parse_robot_name(robot_name)
            if len(cmd_list) > 0:
                if cmd_list[0][0] == 'wait':
                    self.release_at = cmd_list[0][1]

    def update(self):
        channel = super().update()
        handler = getattr(self, "on_" + channel, None)
        if handler is not None:
            handler(getattr(self, channel))
        else:
            assert False, channel  # not supported

# vim: expandtab sw=4 ts=4



"""
  Remove some streams from logfile
"""
from osgar.logger import LogReader, LogWriter, lookup_stream_id, lookup_stream_names


def strip_logfile(logfile, only_stream, outfile):
    if outfile is None:
        outfile = logfile[:-4] + '-strip' + logfile[-4:]
    print('Out:', outfile)
    with LogReader(logfile, only_stream_id=only_stream) as log:
        with LogWriter(filename=outfile, start_time=log.start_time) as out:
            for timestamp, stream_id, data in log:
                out.write(stream_id=stream_id, data=data, dt=timestamp)


def main():
    import argparse

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('logfile', help='input log file')
    parser.add_argument('--keep', help='stream ID or names to keep', nargs='+')
    parser.add_argument('--remove', help='stream ID or names to be removed', nargs='+')
    parser.add_argument('--out', '-o', help='output logfile')
    args = parser.parse_args()

    if args.keep is None and args.remove is None:
        only_stream = None
    else:
        only_stream = [0, ]
        if args.keep is not None:
            names = args.keep
        elif args.remove is not None:
            names = set(lookup_stream_names(args.logfile)) - set(args.remove)
        for name in names:
            only_stream.append(lookup_stream_id(args.logfile, name))

    strip_logfile(args.logfile, only_stream, args.out)


if __name__ == "__main__":
    main()

# vim: expandtab sw=4 ts=4 




"""
  Route with geo2plan and plan2geo conversions + snap position.
"""

import math
from osgar.lib.line import Line, distance, pointAtPolyLineDist


EARTHRADIUS = 6.380935276e+06
SUIDARHTRAE = 1.567168380e-07
DEGRAD            = 1.7453292519943295769236907684886e-02
RADDEG            = 5.7295779513082320876798154814105e+01


def loadLatLonPtsFromStr(input_str):
    lines = input_str.split("\n")
    pts = []
    for line in lines:
        if len(line) > 0 and line[0] != '#':
            assert( len(line.split()) == 2 )
            lat, lon = line.split()
            pts.append( (float(lon), float(lat)) )
    return pts


def loadLatLonPts(filename):
    with open(filename, "r") as f:
        return loadLatLonPtsFromStr(f.read())


def saveLatLonPts(pts, filename, note = None):
    with open(filename, "w") as f:
        if note:
            f.write("# " + note + "\n")
        for p in pts:
            f.write("%f %f\n" % (p[1], p[0]))


class Convertor:
    "convert lat/lon and planar coordinates"
    def __init__(self, refPoint = (16.60906, 49.2060633333)):
        self._cosMulti = max(0.001, EARTHRADIUS*math.cos(refPoint[1]*DEGRAD));
        self.refPoint = refPoint

    def geo2planar(self, pos):
        return ((pos[0] - self.refPoint[0]) * DEGRAD * self._cosMulti, (pos[1] - self.refPoint[1]) * DEGRAD * EARTHRADIUS)

    def planar2geo(self, pos):
        return (self.refPoint[0] + pos[0] * RADDEG/ self._cosMulti, self.refPoint[1] + pos[1] * RADDEG * SUIDARHTRAE)


class DummyConvertor:
    "convert 1:1 for coordinates already in meters"
    def geo2planar(self, pos):
        return pos

    def planar2geo(self, pos):
        return pos


class Route:
    """ route for robot navigation"""
    def __init__(self, pts = [], conv = None, isLoop = None):
        if conv == None:
            conv = Convertor()
        self.conv = conv
        self.pts = [ self.conv.geo2planar(p) for p in pts ]
        if isLoop == None and len(self.pts) > 1:
            isLoop = (distance(self.pts[0], self.pts[-1]) < 5.0)
        self.isLoop = isLoop

    def length( self ):
        if not self.pts:
            return None
        sum = 0
        prev = self.pts[0]
        for p in self.pts:
            sum += distance( prev, p )
            prev = p
        return sum

    def findNearestEx( self, pos ):
        """
        Find nearest position on the polyline and return tuple:
           - nearest point(x,y) on route
           - distance to that point
           - index of segment (negative) or waypoint index (0..N-1)
        """
        if not self.pts:
            return None
        ref = self.conv.geo2planar( pos )
        index = 0
        best = prev = self.pts[0]
        bestDist = distance( ref, best )
        bestIndex = 0
        for p in self.pts[1:]:
            pos, dist, type = Line(prev, p).nearest( ref )
            if dist < bestDist:
                bestDist = dist
                best = pos
                if type < 0:
                    bestIndex = -index -1
                else:
                    bestIndex = index + type
            prev = p
            index += 1
        return self.conv.planar2geo( best ), bestDist, bestIndex

    def findNearest( self, pos ):
        """ filtered version, returning only (x,y) coordinate of nearest point on route """
        near = self.findNearestEx( pos )
        if near:
            return near[0] # position
        return None

    def routeSplit( self, pos ):
        """
        Split path into two parts based on position `pos`:
          - already passed route
          - remaing part to be driven
        """
        snap, dist, index = self.findNearestEx( pos )
        if index < 0:
            # inside line -> split line
            first = [self.conv.planar2geo(p) for p in self.pts[:-index]] + [snap]
            second = [snap]+[self.conv.planar2geo(p) for p in self.pts[-index:]]
        else:
            # some end point
            first = [self.conv.planar2geo(p) for p in self.pts[:index+1]] 
            second = [self.conv.planar2geo(p) for p in self.pts[index:]]
        return first, second

    def pointAtDist( self, dist ):
        """ return point on route in given distance from the start """
        if len(self.pts) < 1:
            return None
        return self.conv.planar2geo( pointAtPolyLineDist( self.pts, dist ) )

    def turnAngleAt( self, pos, radius = 2.0 ):
        """
        Compute "derivation" at nearest point on the route. The radius
        defines how far previous and next point are placed. The returned
        angle is given as a difference from previus and next vectors.
        """
        first, second = self.routeSplit( pos )
        assert len(first) >= 1, len(first)
        assert len(second) >= 1, len(second)
        if self.isLoop:
            geoPts = [self.conv.planar2geo(x) for x in self.pts]
            first = geoPts + first
            second = second + geoPts
        if len(first) == 1 or len(second) == 1:
            return 0.0 # start or end of route
        nextPos = pointAtPolyLineDist( [ self.conv.geo2planar(x) for x in second ], radius )
        currPos = self.conv.geo2planar( second[0] )
        first.reverse()
        prevPos = pointAtPolyLineDist( [ self.conv.geo2planar(x) for x in first ], radius )
        toNext = math.atan2( nextPos[1]-currPos[1], nextPos[0]-currPos[0] )
        toPrev = math.atan2( currPos[1]-prevPos[1], currPos[0]-prevPos[0] )
        ret = toNext - toPrev
        if ret < -math.pi:
            ret += 2*math.pi
        if ret > math.pi:
            ret -= 2*math.pi
        return ret

# vim: expandtab sw=4 ts=4



import unittest
from unittest.mock import MagicMock

from osgar.followme import FollowMe, EmergencyStopException
from osgar.bus import Bus


class FollowMeTest(unittest.TestCase):

    def test_usage(self):
        bus = Bus(MagicMock())
        app = FollowMe(config={}, bus=bus.handle('app'))
        tester = bus.handle('tester')
        tester.register('emergency_stop')
        bus.connect('tester.emergency_stop', 'app.emergency_stop')
        tester.publish('emergency_stop', True)

        app.raise_exception_on_stop = True
        
        with self.assertRaises(EmergencyStopException):
            app.followme()

# vim: expandtab sw=4 ts=4



import math
from osgar.lib.unittest import TestCase

from . import quaternion


class QuaternionTest(TestCase):

    def test_multiply(self):
        qa = [0, 0, 0, 1]
        qb = [1, 0, 0, 0]
        qc = quaternion.multiply(qa, qb)
        self.assertEqual(qc, qb)  # identity

        qd = quaternion.multiply(qb, qb)
        self.assertEqual(qd, [0, 0, 0, -1])  # twice rotation by 90deg

        qdd = quaternion.multiply(qd, qd)
        self.assertEqual(qdd, qa)  # identity

    def test_rotate_z90(self):
        # rotate around Z axis by 90 deg (change heading)
        v = [1, 0, 0]
        q = [ 0, 0, 0.7071068, 0.7071068 ]
        actual = quaternion.rotate_vector(v, q)
        expected = [0, 1, 0]
        self.assertListAlmostEqual(expected, actual, delta=1e-6)

        actual = quaternion.rotate_vector(expected, q)
        expected = [-1, 0, 0]
        self.assertListAlmostEqual(expected, actual, delta=1e-6)

    def test_from_axis_angle(self):
        actual = quaternion.from_axis_angle((0,0,1), math.radians(90))
        expected = [0, 0, 0.7071068, 0.7071068]
        self.assertListAlmostEqual(expected, actual)

    def test_quaterion_internal_normalization(self):
        q = (0.02089575225593203, -0.6597655110187111, 0.381406681115162, -0.7726979260818011)
        angles = quaternion.euler_zyx(q)
        expected = [-1.5436762, 1.0194396, -1.0505728]
        self.assertListAlmostEqual(angles, expected)

    def test_rotation_matrix(self):
        from collections import namedtuple
        T = namedtuple('T', ('q', 'matrix'))
        tests = [
            T([ 0, 0, 0.7071068, 0.7071068 ], [[0.0, -1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, 1.0]]),
            T([ 0, 0.7071068, 0, 0.7071068 ], [[0.0, 0.0, 1.0], [0.0,  1.0,  0.0], [-1.0, 0.0, 0.0]]),
            T([ 0.7071068, 0, 0, 0.7071068 ], [[1.0,  0.0,  0.0], [0.0,  0.0, -1.0], [0.0, 1.0, 0.0 ]]),
            T([ 0.5, 0.5, 0, 0.7071068 ], [[0.5, 0.5, 0.7071068], [0.5, 0.5, -0.7071068], [-0.7071068, 0.7071068, 0.0]]),
            T([ 0, 0, 1, 0 ], [[-1.0, -0.0,  0.0], [0.0000000, -1.0,  0.0], [0.0,  0.0,  1.0]]),
        ]
        for i, t in enumerate(tests):
            with self.subTest(test=i, kind="to matrix"):
                actual = quaternion.rotation_matrix(t.q)
                for row in range(len(actual)):
                    for col in range(len(actual[0])):
                        self.assertAlmostEqual(t.matrix[row][col], actual[row][col], places=6, msg=(row, col))

        for i, t in enumerate(tests):
            with self.subTest(test=i, kind="from matrix"):
                q = quaternion.from_rotation_matrix(t.matrix)
                self.assertListAlmostEqual(q, t.q)


    def test_rotation_matrix_random(self):
        from random import Random
        r = Random(0).random
        for i in range(1000):
            with self.subTest(test=i):
                q = quaternion.normalize([r(), r(), r(), r()])
                rotmat = quaternion.rotation_matrix(q)
                qq = quaternion.from_rotation_matrix(rotmat)
                self.assertListAlmostEqual(q, qq)

# vim: expandtab sw=4 ts=4



import logging
import platform
import signal
import subprocess

g_logger = logging.getLogger(__name__)

class Launch:
    def __init__(self, config, bus):
        self.bus = bus
        self.command = config['command']
        self.shell = config.get('shell', False)

    def start(self):
        self.running = subprocess.Popen(self.command, shell=self.shell)

    def request_stop(self):
        self.bus.shutdown()
        signum = signal.CTRL_C_EVENT if platform.system() == "Windows" else signal.SIGINT
        self.running.send_signal(signum)

    def join(self, timeout=None):
        try:
            self.running.wait(timeout)
        except subprocess.TimeoutExpired:
            command = self.command if isinstance(self.command, str) else " ".join(self.command)
            g_logger.warning(f"'{command}' still running, terminating")
            self.running.terminate()
            try:
                self.running.wait(1)
            except subprocess.TimeoutExpired:
                g_logger.warning(f"'{command}' ignoring SIGTERM, killing")
                self.running.kill()
                self.running.wait()
        assert self.running.poll() is not None


if __name__ == "__main__":
    from unittest.mock import MagicMock
    test_launch = Launch(config={'command': ['echo', 'hello world!']}, bus=MagicMock())
    test_launch.start()
    test_launch.join()

    test_request_stop = Launch(config={'command': ['sleep', '10']}, bus=MagicMock())
    test_request_stop.start()
    test_request_stop.request_stop()
    test_request_stop.join(0.1)

    test_join = Launch(config={'command': ['sleep', '10']}, bus=MagicMock())
    test_join.start()
    test_join.join(0.1)



import unittest
from unittest.mock import patch, MagicMock
from datetime import timedelta
import json
import os
import threading
import signal
import subprocess
import tempfile
import sys

from osgar.record import Recorder


class Sleeper:

    def __init__(self, cfg, bus):
        self.e = threading.Event()

    def start(self):
        self.t = threading.Thread(target=self.e.wait, args=(5,))
        self.t.start()

    def join(self, timeout=None):
        self.t.join(timeout)

    def request_stop(self):
        self.e.set()


class RecorderTest(unittest.TestCase):

    def test_dummy_usage(self):
        empty_config = {'modules': {}, 'links':[]}
        with Recorder(config=empty_config, logger=MagicMock()) as recorder:
            pass

    def test_missing_init(self):
        # init section for modules is now optional
        mini_config = {'modules': {
            "dummy": {
                "driver": "osgar.test_record:Sleeper"
            },
        }, 'links':[]}
        with Recorder(config=mini_config, logger=MagicMock()) as recorder:
            pass

    def test_config(self):
        with patch('osgar.drivers.logserial.serial.Serial') as mock:
            instance = mock.return_value
            instance.read = MagicMock(return_value=b'$GNGGA,182433.10,5007.71882,N,01422.50467,E,1,05,6.09,305.1,M,44.3,M,,*41')

            config = {
                    'modules': {
                        'gps': {
                            'driver': 'gps',
                            'out':['position'],
                            'init':{}
                        },
                        'serial_gps': {
                            'driver': 'serial',
                            'out':['raw'],
                            'init': {'port': 'COM51', 'speed': 4800}
                        }
                    },
                    'links': [('serial_gps.raw', 'gps.raw')]
            }
            logger = MagicMock(write = MagicMock(return_value=timedelta(seconds=135)))
            with Recorder(config=config, logger=logger) as recorder:
                self.assertEqual(len(recorder.modules), 2)
                self.assertEqual(sum([sum([len(q) for q in module.bus.out.values()])
                                      for module in recorder.modules.values()]), 1)

    def test_spider_config(self):
        # first example with loop spider <-> serial
        with open(os.path.dirname(__file__) + '/../config/test-spider.json') as f:
            config = json.loads(f.read())

        with patch('osgar.drivers.logserial.serial.Serial') as mock:
            logger = MagicMock()
            recorder = Recorder(config=config['robot'], logger=logger)


    def test_all_supported_config_files(self):
        supported = ['test-spider.json', 'test-gps-imu.json',
                'test-spider-gps-imu.json', 'test-windows-gps.json']

        with patch('osgar.drivers.logserial.serial.Serial') as mock:
            logger = MagicMock()
            for filename in supported:
                with open(os.path.join(os.path.dirname(__file__), '..', 'config',
                                   filename)) as f:
                    config = json.loads(f.read())
                recorder = Recorder(config=config['robot'], logger=logger)

    @unittest.skipIf(os.name != "posix", "requires posix shell")
    def test_sigint_shell(self):
        config = {
            'version': 2,
            'robot': {
                'modules': {
                    "app": {
                        "driver": "osgar.test_record:Sleeper",
                        "init": {}
                    },
                }, 'links':[]
            }
        }
        with tempfile.NamedTemporaryFile() as cfg:
            cfg.write(json.dumps(config).encode('ascii'))
            cfg.flush()
            env = os.environ.copy()
            env['OSGAR_LOGS'] = '.'
            with subprocess.Popen(
                    f"echo starting; {sys.executable} -m osgar.record {cfg.name}; echo should not get here",
                    shell=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    start_new_session=True,
                    env=env,
            ) as proc:
                grp_id = os.getpgid(proc.pid)
                self.assertEqual(proc.stdout.readline().strip(), b"starting")
                log_line = proc.stderr.readline().strip().split()
                log_filename = log_line[-1]
                self.assertTrue(log_filename.endswith(b".log"), log_line)
                self.assertIn(b"SIGINT handler installed", proc.stderr.readline())
                os.killpg(grp_id, signal.SIGINT)
                stdout, stderr = proc.communicate()
                self.assertIn(b"committing suicide by SIGINT", stderr)
                self.assertEqual(len(stdout), 0, stdout)
                self.assertEqual(len(stderr.splitlines()), 1, stderr)
            os.unlink(log_filename)


# vim: expandtab sw=4 ts=4



"""
  Extract position with topics statistic from AWS ROS_INFO output
"""


def aws2info(filename, outname):
    KEY = 'Python3: stdout '
    with open(filename) as f, open(outname, 'w') as out:
        for line in f:
            if KEY in line and line.strip().endswith(']'):
                out.write(line[line.index(KEY)+len(KEY):])


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('filename', help='AWS ROS recorded log file')
    args = parser.parse_args()

    assert 'rosout.log' in args.filename, args.filename
    outname = args.filename.replace('rosout.log', 'info.txt')
    aws2info(args.filename, outname=outname)

# vim: expandtab sw=4 ts=4



import numpy as np
import torch

import subt.artf_model


class Detector:
    def __init__(self, model, confidence_thresholds, categories, device,
                 max_gap, min_group_size):
        self.model = model
        self.confidence_thresholds = confidence_thresholds
        self.categories = categories
        self.device = device
        self.max_gap = max_gap
        self.min_group_size = min_group_size

    def __call__(self, img):
        with torch.set_grad_enabled(False):
            timg = torch.from_numpy(img).float().to(self.device)
            # We work with a batch of a single image.
            minibatch = timg.unsqueeze(0)
            predictions = np.ascontiguousarray(
                self.model(minibatch)[0].detach().cpu().numpy())
        scale = img.shape[0] // predictions.shape[0]
        indices = np.indices(predictions.shape[:2]) * scale

        artifacts = []
        for name, aid in self.categories.items():
            threshold = self.confidence_thresholds.get(name)
            if threshold is None:
                continue
            detected = predictions[:, :, aid] >= threshold
            ys = indices[0, detected].reshape((-1, ))
            xs = indices[1, detected].reshape((-1, ))
            for (x, y) in zip(xs, ys):
                confidence = predictions[y // scale, x // scale, aid]
                artifacts.append((x, y, name, confidence))
        return Detector.prune_small_groups(
            Detector.group_points(artifacts, self.max_gap),
            self.min_group_size)

    @staticmethod
    def group_points(artifacts, max_gap):
        belongs_to = dict()  # Maps from points to a group they belong to.
        consists_of = dict()  # Maps from group ids to a list of points.
        next_group_id = 0
        X = 0
        Y = 1
        CATEGORY = 2
        CONFIDENCE = 3
        for i in range(len(artifacts)):
            artf_i = artifacts[i]
            group_i = next_group_id
            next_group_id += 1
            belongs_to[i] = group_i
            consists_of[group_i] = [i]
            category_i = artf_i[CATEGORY]
            for j in range(i):
                artf_j = artifacts[j]
                category_j = artf_j[CATEGORY]
                dist = np.hypot(artf_i[X] - artf_j[X], artf_i[Y] - artf_j[Y])
                if dist <= max_gap and category_i == category_j:
                    group_j = belongs_to[j]
                    if group_i == group_j:
                        # The two points allready belong to the same group and
                        # there is nothing to be done.
                        continue
                    else:
                        # We need to merge two groups.
                        # We are merging a smaller group into the bigger one to
                        # pretend we care about speed. For the small groups we
                        # expect, this probably does not make much difference.
                        if (len(consists_of[group_j]) >= len(
                                consists_of[group_i])):
                            merge_to_group = group_j
                            merge_from_group = group_i
                        else:
                            merge_to_group = group_i
                            merge_from_group = group_j
                        for k in consists_of[merge_from_group]:
                            belongs_to[k] = merge_to_group
                            consists_of[merge_to_group].append(k)
                        del consists_of[merge_from_group]
                        group_i = merge_to_group

        groups = []
        for group in consists_of.values():
            # We want the most representative point in the group to go first,
            # because the neural network sometimes classifies points of the
            # object to still belong to it. In a 3D scene, these points have a
            # very different coordinate and would lead to errors!
            cx = np.mean([artifacts[i][X] for i in group])
            cy = np.mean([artifacts[i][Y] for i in group])
            group.sort(key=lambda i: np.hypot(artifacts[i][X] - cx, artifacts[
                i][Y] - cy))
            category = artifacts[group[0]][CATEGORY]
            groups.append([category, [[artifacts[i][X].item(), artifacts[i][Y].item(),
                                       artifacts[i][CONFIDENCE].item()]
                                      for i in group]])
        return groups

    @staticmethod
    def prune_small_groups(groups, min_group_size):
        POINTS = 1
        return [
            group for group in groups if len(group[POINTS]) >= min_group_size
        ]


def thresholds_param(arg):
    thresholds = {}
    parts = arg.split(',')
    for part in parts:
        subparts = part.split(':')
        thresholds[subparts[0]] = float(subparts[1])
    return thresholds


if __name__ == '__main__':
    import argparse
    import cv2
    import os
    import sys

    from osgar.lib.serialize import deserialize
    from osgar.logger import LogReader, lookup_stream_id

    parser = argparse.ArgumentParser()
    parser.add_argument('--model',
                        default=None,
                        help='Path to model parameters. Example: ' +
                        'mdnet0.64.64.13.4.relu.pth')
    parser.add_argument('--logfile', help='Path to Osgar log.')
    parser.add_argument('--image-stream',
                        help='Stream id or name with image data.')
    parser.add_argument('--image', help='An image to process.')
    parser.add_argument('--output', help='Where to store detection images.')
    parser.add_argument('--confidence-thresholds',
                        type=thresholds_param,
                        default={},
                        help='Minimum confidence needed to declare a ' +
                        'detected artifact. Example: survivor:0.95,' +
                        'vent:0.878,backpack:0.977,phone:0.97,' +
                        'helmet:0.963,rope:0.99')
    parser.add_argument(
        '--max-gap',
        type=int,
        default=16,
        help='Maximum gap, in pixels, between two points that' +
        'we are still willing to consider to belong to ' +
        'the same artifact.')
    parser.add_argument('--min-group-size',
                        type=int,
                        default=2,
                        help='There need to be at least this many points in ' +
                        'an artifact for it to be considered real and ' +
                        'not noise.')
    parser.add_argument(
        '--show',
        default=False,
        action='store_true',
        help='Enable visualization of the scene and detected ' + 'artifacts.')
    args = parser.parse_args()
    assert (args.model)
    assert ((args.logfile and args.image_stream) or (args.image))

    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda" if use_cuda else "cpu")
    print('Using:', device, file=sys.stderr)
    model, categories = subt.artf_model.load_model(args.model, device)
    detector = Detector(model, args.confidence_thresholds, categories, device,
                        args.max_gap, args.min_group_size)

    n = 0
    if args.logfile:
        image_stream_id = lookup_stream_id(args.logfile, args.image_stream)
        with LogReader(args.logfile,
                       only_stream_id=image_stream_id) as logreader:
            for time, stream, data in logreader:
                assert (stream == image_stream_id)
                img = cv2.imdecode(np.frombuffer(deserialize(data), np.uint8),
                                   cv2.IMREAD_COLOR)
                artifacts = detector(img)
                print('---')
                for name, group in artifacts:
                    print(name)
                    for x, y, confidence in group:
                        print(' ', confidence, x, y)
                    if args.show or args.output:
                        min_x = min(x for (x, _, _) in group)
                        min_y = min(y for (_, y, _) in group)
                        max_x = max(x for (x, _, _) in group)
                        max_y = max(y for (_, y, _) in group)
                        # Rectangle around all detected points.
                        cv2.rectangle(img, (min_x, min_y), (max_x, max_y), (0, 255, 0))
                        # Main point.
                        cv2.circle(img, group[0][:2], 5, (0, 0, 255), -1)
                if artifacts and args.output:
                    cv2.imwrite(
                        os.path.join(args.output, '{:06d}.jpg'.format(n)),
                        img)
                n += 1
                if args.show:
                    cv2.imshow('img', img)
                    KEY_Q = ord('q')
                    if artifacts:
                        key = cv2.waitKey() & 0xFF
                    else:
                        key = cv2.waitKey(3) & 0xFF
                    if key == KEY_Q:
                        break

    if args.image:
        img = cv2.imread(args.image, cv2.IMREAD_COLOR)
        artifacts = detector(img)
        for name, group in artifacts:
            print(name)
            for x, y, confidence in group:
                print('  ', confidence, x, y)
            if args.show or args.output:
                min_x = min(x for (x, _, _) in group)
                min_y = min(y for (_, y, _) in group)
                max_x = max(x for (x, _, _) in group)
                max_y = max(y for (_, y, _) in group)
                # Rectangle around all detected points.
                cv2.rectangle(img, (min_x, min_y), (max_x, max_y), (0, 255, 0))
                # Main point.
                cv2.circle(img, group[0][:2], 5, (0, 0, 255), -1)
        if args.output:
            cv2.imwrite(
                os.path.join(args.output, os.path.basename(args.image)), img)
        if args.show:
            cv2.imshow('img', img)
            cv2.waitKey()



"""
  SubT Challenge Version 1
"""
import gc
import os.path
import math
import threading
import copy

from datetime import timedelta
from collections import defaultdict
from io import StringIO

import numpy as np

from osgar.explore import follow_wall_angle
from osgar.lib.mathex import normalizeAnglePIPI
from osgar.lib import quaternion
from osgar.lib.virtual_bumper import VirtualBumper
from osgar.lib.lidar_pts import equal_scans
from osgar.lib.loop import LoopDetector

from subt.local_planner import LocalPlanner
from subt.trace import Trace, distance3D
from subt.name_decoder import parse_robot_name

# safety limits for exploration and return home
DEFAULT_LIMIT_ROLL = 28  # degrees
DEFAULT_LIMIT_PITCH = 28  # degrees
DEFAULT_RETURN_LIMIT_ROLL = 35  # degrees
DEFAULT_RETURN_LIMIT_PITCH = 35  # degrees

# accepted LoRa commands
LORA_GO_HOME_CMD = b'GoHome'
LORA_STOP_CMD = b'Stop'
LORA_PAUSE_CMD = b'Pause'
LORA_CONTINUE_CMD = b'Continue'

# reasons for termination of follow wall
REASON_DIST_REACHED = 'dist_reached'
REASON_PITCH_LIMIT = 'pitch_limit'
REASON_ROLL_LIMIT = 'roll_limit'
REASON_VIRTUAL_BUMPER = 'virtual_bumper'
REASON_LORA = 'lora'
REASON_FRONT_BUMPER = 'front_bumper'
REASON_REAR_BUMPER = 'rear_bumper'
REASON_LOOP = 'loop'


def any_is_none(*args):
    for a in args:
        if a is None:
            return True
    return False


def min_dist(laser_data):
    if len(laser_data) > 0:
        # remove ultra near reflections and unlimited values == 0
        laser_data = [x if x > 10 else 10000 for x in laser_data]
        return min(laser_data)/1000.0
    return 0


def distance(pose1, pose2):
    return math.hypot(pose1[0] - pose2[0], pose1[1] - pose2[1])


class Collision(Exception):
    pass


class EmergencyStopException(Exception):
    pass


class EmergencyStopMonitor:
    def __init__(self, robot):
        self.robot = robot

    def update(self, robot):
        if robot.emergency_stop:
            raise EmergencyStopException()

        # handle STOP independently of current subroutine
        if robot.lora_cmd == LORA_STOP_CMD:
            print(robot.time, 'LoRa cmd - Stop')
            raise EmergencyStopException()

    # context manager functions
    def __enter__(self):
        self.callback = self.robot.register(self.update)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.robot.unregister(self.callback)


class SubTChallenge:
    def __init__(self, config, bus):
        self.bus = bus
        bus.register("desired_speed", "pose2d", "artf_xyz", "stdout", "request_origin", "desired_z_speed")
        self.traveled_dist = 0.0
        self.time = None
        self.max_speed = config['max_speed']
        self.max_angular_speed = math.radians(60)
        self.walldist = config['walldist']
        self.timeout = timedelta(seconds=config['timeout'])
        self.symmetric = config['symmetric']  # is robot symmetric?
        self.dangerous_dist = config.get('dangerous_dist', 0.3)
        self.min_safe_dist = config.get('min_safe_dist', 0.75)
        self.safety_turning_coeff = config.get('safety_turning_coeff', 0.8)
        self.limit_pitch = math.radians(config.get('limit_pitch', DEFAULT_LIMIT_PITCH))
        self.limit_roll = math.radians(config.get('limit_roll', DEFAULT_LIMIT_ROLL))
        self.return_limit_pitch = math.radians(config.get('return_limit_pitch', DEFAULT_RETURN_LIMIT_PITCH))
        self.return_limit_roll = math.radians(config.get('return_limit_roll', DEFAULT_RETURN_LIMIT_ROLL))
        virtual_bumper_sec = config.get('virtual_bumper_sec')
        self.virtual_bumper = None
        if virtual_bumper_sec is not None:
            virtual_bumper_radius = config.get('virtual_bumper_radius', 0.1)
            self.virtual_bumper = VirtualBumper(timedelta(seconds=virtual_bumper_sec), virtual_bumper_radius)
        self.speed_policy = config.get('speed_policy', 'always_forward')
        assert(self.speed_policy in ['always_forward', 'conservative'])
        self.height_above_ground = config.get('height_above_ground', 0.0)
        self.trace_z_weight = config.get('trace_z_weight', 0.2)  # Z is important for drones ~ 3.0

        self.last_position = (0, 0, 0)  # proper should be None, but we really start from zero
        self.xyz = None  # 3D position for mapping artifacts - unknown depends on source (pose2d or pose3d)
        self.yaw, self.pitch, self.roll = 0, 0, 0
        self.orientation = None  # quaternion updated by on_pose3d()
        self.yaw_offset = None  # not defined, use first IMU reading
        self.is_moving = None  # unknown
        self.scan = None  # I should use class Node instead
        self.flipped = False  # by default use only front part
        self.joint_angle_rad = []  # optinal angles, needed for articulated robots flip
        self.stat = defaultdict(int)
        self.voltage = []
        self.artifacts = []
        self.trace = Trace()
        self.loop_detector = LoopDetector()
        self.collision_detector_enabled = False
        self.sim_time_sec = 0

        self.lora_cmd = None

        self.emergency_stop = None
        self.monitors = []  # for Emergency Stop Exception

        self.use_right_wall = config['right_wall']
        self.use_center = False  # navigate into center area (controlled by name ending by 'C')
        self.is_virtual = config.get('virtual_world', False)  # workaround to handle tunnel differences

        self.front_bumper = False
        self.rear_bumper = False

        self.last_send_time = None
        self.origin = None  # unknown initial position
        self.origin_quat = quaternion.identity()

        self.init_path = None
        if 'init_path' in config:
            pts_s = [s.split(',') for s in config['init_path'].split(';')]
            self.init_path = [(float(x), float(y)) for x, y in pts_s]
        self.origin_error = False
        self.robot_name = None  # received with origin
        scan_subsample = config.get('scan_subsample', 1)
        obstacle_influence = config.get('obstacle_influence', 0.8)
        direction_adherence = math.radians(config.get('direction_adherence', 90))
        max_obstacle_distance = config.get('max_obstacle_distance', 2.5)
        self.local_planner = LocalPlanner(
                obstacle_influence=obstacle_influence,
                direction_adherence=direction_adherence,
                max_obstacle_distance=max_obstacle_distance,
                scan_subsample=scan_subsample,
                max_considered_obstacles=100)
        self.use_return_trace = config.get('use_return_trace', True)
        self.ref_scan = None
        self.pause_start_time = None
        if config.get('start_paused', False):
            self.pause_start_time = timedelta()  # paused from the very beginning

    def send_speed_cmd(self, speed, angular_speed):
        if self.virtual_bumper is not None:
            self.virtual_bumper.update_desired_speed(speed, angular_speed)
        self.bus.publish('desired_speed', [round(speed*1000), round(math.degrees(angular_speed)*100)])
        # Corresponds to gc.disable() in __main__. See a comment there for more details.
        gc.collect()

    def maybe_remember_artifact(self, artifact_data, artifact_xyz):
        for stored_data, (x, y, z) in self.artifacts:
            if distance3D((x, y, z), artifact_xyz) < 4.0:
                # in case of uncertain type, rather report both
                if stored_data == artifact_data:
                    return False
        self.artifacts.append((artifact_data, artifact_xyz))
        return True

    def go_straight(self, how_far, timeout=None):
        print(self.time, "go_straight %.1f (speed: %.1f)" % (how_far, self.max_speed), self.last_position)
        start_pose = self.last_position
        if how_far >= 0:
            self.send_speed_cmd(self.max_speed, 0.0)
        else:
            self.send_speed_cmd(-self.max_speed, 0.0)
        start_time = self.time
        while distance(start_pose, self.last_position) < abs(how_far):
            self.update()
            if timeout is not None and self.time - start_time > timeout:
                print("go_straight - TIMEOUT!")
                break
        self.send_speed_cmd(0.0, 0.0)

    def go_safely(self, desired_direction):
        if self.local_planner is None:
            safety, safe_direction = 1.0, desired_direction
        else:
            safety, safe_direction = self.local_planner.recommend(desired_direction)
        #print(self.time,"safety:%f    desired:%f  safe_direction:%f"%(safety, desired_direction, safe_direction))
        if self.speed_policy == 'conservative':
            desired_angular_speed = 1.2 * safe_direction
        elif self.speed_policy == 'always_forward':
            desired_angular_speed = 0.9 * safe_direction
        else:
            assert(False)  # Unsupported speed_policy.
        size = len(self.scan)
        dist = min_dist(self.scan[size//3:2*size//3])
        if dist < self.min_safe_dist:  # 2.0:
            if self.speed_policy == 'conservative':
                desired_speed = self.max_speed * (1.2/2.0) * (dist - 0.4) / 1.6
            elif self.speed_policy == 'always_forward':
                desired_speed = self.max_speed * (dist - self.dangerous_dist) / (self.min_safe_dist - self.dangerous_dist)
            else:
                assert(False)  # Unsupported speed policy.
        else:
            desired_speed = self.max_speed  # was 2.0
        desired_speed = desired_speed * (1.0 - self.safety_turning_coeff * min(self.max_angular_speed, abs(desired_angular_speed)) / self.max_angular_speed)
        if self.flipped:
            self.send_speed_cmd(-desired_speed, desired_angular_speed)  # ??? angular too??!
        else:
            self.send_speed_cmd(desired_speed, desired_angular_speed)
        return safety

    def turn(self, angle, with_stop=True, speed=0.0, timeout=None):
        print(self.time, "turn %.1f" % math.degrees(angle))
        start_pose = self.last_position
        if angle >= 0:
            self.send_speed_cmd(speed, self.max_angular_speed)
        else:
            self.send_speed_cmd(speed, -self.max_angular_speed)
        start_time = self.time
        while abs(normalizeAnglePIPI(start_pose[2] - self.last_position[2])) < abs(angle):
            self.update()
            if timeout is not None and self.time - start_time > timeout:
                print(self.time, "turn - TIMEOUT!")
                break
        if with_stop:
            self.send_speed_cmd(0.0, 0.0)
            start_time = self.time
            while self.time - start_time < timedelta(seconds=2):
                self.update()
                if not self.is_moving:
                    break
            print(self.time, 'stop at', self.time - start_time)

    def stop(self):
        self.send_speed_cmd(0.0, 0.0)
        start_time = self.time
        while self.time - start_time < timedelta(seconds=20):
            self.update()
            if not self.is_moving:
                break
        print(self.time, 'stop at', self.time - start_time, self.is_moving)

    def follow_wall(self, radius, right_wall=False, timeout=timedelta(hours=3), dist_limit=None, flipped=False,
            pitch_limit=None, roll_limit=None, detect_loop=True):
        # make sure that we will start with clean data
        if flipped:
            self.scan = None
            self.flipped = True

        reason = None  # termination reason is not defined yet
        start_dist = self.traveled_dist
        start_time = self.sim_time_sec

        last_pause_time = timedelta()  # for multiple Pause
        current_pause_time = timedelta()
        while self.sim_time_sec - start_time < (timeout + last_pause_time + current_pause_time).total_seconds():
            try:
                channel = self.update()
                if (channel == 'scan' and not self.flipped) or (channel == 'scan_back' and self.flipped) or channel == 'scan360':
                    if self.pause_start_time is None:
                        if self.use_center:
                            desired_direction = 0
                        else:
                            desired_direction = follow_wall_angle(self.scan, radius=radius, right_wall=right_wall)
                        self.go_safely(desired_direction)
                if dist_limit is not None:
                    if dist_limit < abs(self.traveled_dist - start_dist):  # robot can return backward -> abs()
                        print(self.time, 'Distance limit reached! At', self.traveled_dist, self.traveled_dist - start_dist)
                        reason = REASON_DIST_REACHED
                        break
                if pitch_limit is not None and self.pitch is not None:
                    if abs(self.pitch) > pitch_limit:
                        print(self.time, 'Pitch limit triggered termination: (pitch %.1f)' % math.degrees(self.pitch))
                        reason = REASON_PITCH_LIMIT
                        break
                if roll_limit is not None and self.roll is not None:
                    if abs(self.roll) > roll_limit:
                        print(self.time, 'Roll limit triggered termination: (roll %.1f)' % math.degrees(self.roll))
                        reason = REASON_ROLL_LIMIT
                        break
                if self.virtual_bumper is not None and self.virtual_bumper.collision():
                    print(self.time, "VIRTUAL BUMPER - collision")
                    self.go_straight(-0.3, timeout=timedelta(seconds=10))
                    reason = REASON_VIRTUAL_BUMPER
                    break

                loop = detect_loop and self.loop_detector.loop()
                if loop:
                    print(self.time, self.sim_time_sec, 'Loop detected')
                    self.turn(math.radians(160 if self.use_right_wall else -160), timeout=timedelta(seconds=40), with_stop=True)
                    reason = REASON_LOOP
                    break

                if self.front_bumper and not flipped:
                    print(self.time, "FRONT BUMPER - collision")
                    self.go_straight(-0.3, timeout=timedelta(seconds=10))
                    reason = REASON_FRONT_BUMPER
                    break

                if self.rear_bumper and flipped:
                    print(self.time, "REAR BUMPER - collision")
                    self.go_straight(-0.3, timeout=timedelta(seconds=10))
                    reason = REASON_REAR_BUMPER
                    break

                if self.lora_cmd is not None:
                    # the "GoHome" command must be accepted only on the way there and not on the return home
                    if dist_limit is None and self.lora_cmd == LORA_GO_HOME_CMD:
                        print(self.time, 'LoRa cmd - GoHome')
                        self.lora_cmd = None
                        reason = REASON_LORA
                        break
                    if self.lora_cmd == LORA_PAUSE_CMD:
                        print(self.time, 'LoRa cmd - Pause')
                        self.send_speed_cmd(0, 0)
                        if self.pause_start_time is None:
                            # ignore repeated Pause
                            self.pause_start_time = self.time
                        self.lora_cmd = None
                    elif self.lora_cmd == LORA_CONTINUE_CMD:
                        print(self.time, 'LoRa cmd - Continue')
                        if self.pause_start_time is not None:
                            # ignore Continue without Pause
                            last_pause_time += self.time - self.pause_start_time
                            self.pause_start_time = None
                        self.lora_cmd = None
                if self.pause_start_time is not None:
                    current_pause_time = self.time - self.pause_start_time
                else:
                    current_pause_time = timedelta()
            except Collision:
                assert not self.collision_detector_enabled  # collision disables further notification
                before_stop = self.xyz
                self.stop()
                after_stop = self.xyz
                print("Pose Jump:", before_stop, after_stop)
                self.xyz = before_stop
                self.go_straight(-1)
                self.stop()
                if right_wall:
                    turn_angle = math.pi / 2
                else:
                    turn_angle = -math.pi / 2
                self.turn(turn_angle, with_stop=True)
                self.go_straight(1.5)
                self.stop()
                self.turn(-turn_angle, with_stop=True)
                self.go_straight(1.5)
                self.stop()
                self.collision_detector_enabled = True
        self.scan = None
        self.flipped = False
        return self.traveled_dist - start_dist, reason

    def return_home(self, timeout, home_threshold=None):
        if home_threshold is None:
            HOME_THRESHOLD = 5.0
        else:
            HOME_THRESHOLD = home_threshold
        SHORTCUT_RADIUS = 2.3
        MAX_TARGET_DISTANCE = 5.0
        MIN_TARGET_DISTANCE = 1.0
        assert(MAX_TARGET_DISTANCE > SHORTCUT_RADIUS) # Because otherwise we could end up with a target point more distant from home than the robot.
        print('Wait and get ready for return')
        self.send_speed_cmd(0, 0)
        self.wait(dt=timedelta(seconds=3.0))
        original_trace = copy.deepcopy(self.trace)
        self.trace.prune(SHORTCUT_RADIUS)
        self.wait(dt=timedelta(seconds=2.0))
        print('done.')
        home_position = self.trace.start_position()
        if home_position is None:
            return  # empty trace, no start -> we are at home
        start_time = self.sim_time_sec
        target_distance = MAX_TARGET_DISTANCE
        count_down = 0
        while distance3D(self.xyz, home_position) > HOME_THRESHOLD and self.sim_time_sec - start_time < timeout.total_seconds():
            channel = self.update()
            if (channel == 'scan' and not self.flipped) or (channel == 'scan_back' and self.flipped) or (channel == 'scan360'):
                if target_distance == MIN_TARGET_DISTANCE:
                    target_x, target_y, target_z = original_trace.where_to(self.xyz, target_distance, self.trace_z_weight)
                else:
                    target_x, target_y, target_z = self.trace.where_to(self.xyz, target_distance, self.trace_z_weight)
#                print(self.time, self.xyz, (target_x, target_y), math.degrees(self.yaw))
                x, y = self.xyz[:2]
                desired_direction = math.atan2(target_y - y, target_x - x) - self.yaw
                if self.flipped:
                    desired_direction += math.pi  # symmetry
                    for angle in self.joint_angle_rad:
                        desired_direction -= angle

                d = distance3D(self.xyz, [target_x, target_y, target_z])
                time_to_target = d/self.max_speed
                desired_z_speed = (target_z - self.xyz[2]) / time_to_target
                self.bus.publish('desired_z_speed', desired_z_speed)

                safety = self.go_safely(desired_direction)
                if safety < 0.2:
                    print(self.time, "Safety low!", safety, desired_direction)
                    target_distance = MIN_TARGET_DISTANCE
                    count_down = 300
                if count_down > 0:
                    count_down -= 1
                    if count_down == 0:
                        target_distance = MAX_TARGET_DISTANCE
                        print(self.time, "Recovery to original", target_distance)

        print('return_home: dist', distance3D(self.xyz, home_position), 'time(sec)', self.sim_time_sec - start_time)
        self.bus.publish('desired_z_speed', None)

    def follow_trace(self, trace, timeout, max_target_distance=5.0, safety_limit=None):
        print('Follow trace')
        END_THRESHOLD = 2.0
        start_time = self.sim_time_sec
        print('MD', self.xyz, distance3D(self.xyz, trace.trace[0]), trace.trace)
        while distance3D(self.xyz, trace.trace[0]) > END_THRESHOLD and self.sim_time_sec - start_time < timeout.total_seconds():
            if self.update() in ['scan', 'scan360']:
                target_x, target_y = trace.where_to(self.xyz, max_target_distance, self.trace_z_weight)[:2]
                x, y = self.xyz[:2]
                desired_direction = math.atan2(target_y - y, target_x - x) - self.yaw
                safety = self.go_safely(desired_direction)
                if safety_limit is not None and safety < safety_limit:
                    print('Danger! Safety limit for follow trace reached!', safety, safety_limit)
                    break
        print('End of follow trace(sec)', self.sim_time_sec - start_time)

    def register(self, callback):
        self.monitors.append(callback)
        return callback

    def unregister(self, callback):
        assert callback in self.monitors
        self.monitors.remove(callback)

    def on_pose2d(self, timestamp, data):
        if self.xyz is None:
            self.xyz = 0, 0, 0
        x, y, heading = data
        pose = (x / 1000.0, y / 1000.0, math.radians(heading / 100.0))
        if self.last_position is not None:
            self.is_moving = (self.last_position != pose)
            dist = math.hypot(pose[0] - self.last_position[0], pose[1] - self.last_position[1])
            direction = ((pose[0] - self.last_position[0]) * math.cos(self.last_position[2]) +
                         (pose[1] - self.last_position[1]) * math.sin(self.last_position[2]))
            if direction < 0:
                dist = -dist
        else:
            dist = 0.0
        self.last_position = pose
        self.traveled_dist += dist
        x, y, z = self.xyz
        x += math.cos(self.pitch) * math.cos(self.yaw) * dist
        y += math.cos(self.pitch) * math.sin(self.yaw) * dist
        z += math.sin(self.pitch) * dist
        self.last_send_time = self.bus.publish('pose2d', [round(x * 1000), round(y * 1000),
                                    round(math.degrees(self.yaw) * 100)])
        if self.virtual_bumper is not None:
            if self.is_virtual:
                self.virtual_bumper.update_pose(timedelta(seconds=self.sim_time_sec), pose)
            else:
                self.virtual_bumper.update_pose(self.time, pose)
        self.xyz = x, y, z
        self.trace.update_trace(self.xyz)

    def on_pose3d(self, timestamp, data):
        xyz, rot = data
        self.orientation = rot  # quaternion
        ypr = quaternion.euler_zyx(rot)

        pose = (xyz[0], xyz[1], ypr[0])
        if self.last_position is not None:
            self.is_moving = (self.last_position != pose)
            dist = math.hypot(pose[0] - self.last_position[0], pose[1] - self.last_position[1])
            direction = ((pose[0] - self.last_position[0]) * math.cos(self.last_position[2]) +
                         (pose[1] - self.last_position[1]) * math.sin(self.last_position[2]))
            if direction < 0:
                dist = -dist
        else:
            dist = 0.0
        self.last_position = pose
        self.traveled_dist += dist
        x, y, z = xyz
        self.last_send_time = self.bus.publish('pose2d', [round(x * 1000), round(y * 1000),
                                    round(math.degrees(self.yaw) * 100)])
        if self.virtual_bumper is not None:
            if self.is_virtual:
                self.virtual_bumper.update_pose(timedelta(seconds=self.sim_time_sec), pose)
            else:
                self.virtual_bumper.update_pose(self.time, pose)
        self.xyz = tuple(xyz)
        self.trace.update_trace(self.xyz)
        self.loop_detector.add(self.xyz, rot)

    def on_acc(self, timestamp, data):
        acc = [x / 1000.0 for x in data]
        gacc = np.matrix([[0., 0., 9.80]])  # Gravitational acceleration.
        cos_pitch = math.cos(self.pitch)
        sin_pitch = math.sin(self.pitch)
        # TODO: Once roll is correct, incorporate it here too.
        egacc = np.matrix([  # Expected gravitational acceleration given known pitch.
            [cos_pitch, 0., sin_pitch],
            [0., 1., 0.],
            [-sin_pitch, 0., cos_pitch]
        ]) * gacc.T
        cacc = np.asarray(acc) - egacc.T  # Corrected acceleration (without gravitational acceleration).
        magnitude = math.hypot(cacc[0, 0], cacc[0, 1])
        # used to be 12 - see https://bitbucket.org/osrf/subt/issues/166/expected-x2-acceleration
        if magnitude > 200:  #18.0:
            print(self.time, 'Collision!', acc, 'reported:', self.collision_detector_enabled)
            if self.collision_detector_enabled:
                self.collision_detector_enabled = False
                raise Collision()

    def publish_single_artf_xyz(self, artifact_data, pos):
        ax, ay, az = pos
        self.bus.publish('artf_xyz', [
            [artifact_data, [round(ax * 1000), round(ay * 1000), round(az * 1000)], self.robot_name, None]])

    def on_artf(self, timestamp, data):
        if self.orientation is None or self.xyz is None:
            # there can be observed artifact (false) on the start before the coordinate system is defined
            return
        artifact_data, vector = data
        dx, dy, dz = quaternion.rotate_vector(vector, self.orientation)
        x, y, z = self.xyz
        ax = x + dx/1000.0
        ay = y + dy/1000.0
        az = z + dz/1000.0
        if -20 < ax < 0 and -10 < ay < 10:  # AND of currently available staging areas
            # Urban (-20 < ax < 0 and -10 < ay < 10)
            # Cave  (-50 < ax < 0 and -25 < ay < 25)
            # filter out elements on staging area
            self.stdout(self.time, 'Robot at:', (ax, ay, az))
        else:
            if self.maybe_remember_artifact(artifact_data, (ax, ay, az)):
                self.publish_single_artf_xyz(artifact_data, (ax, ay, az))

    def on_joint_angle(self, timestamp, data):
        # angles for articulated robot in 1/100th of degree
        self.joint_angle_rad = [math.radians(a/100) for a in data]

    def on_bumpers_front(self, timestamp, data):
        self.front_bumper = max(data)  # array of boolean values where True means collision

    def on_bumpers_rear(self, timestamp, data):
        self.rear_bumper = max(data)  # array of boolean values where True means collision

    def on_origin(self, timestamp, data):
        if self.origin is None:  # accept only initial offset
            self.robot_name = data[0].decode('ascii')
            if len(data) == 8:
                self.origin = data[1:4]
                qx, qy, qz, qw = data[4:]
                self.origin_quat = qx, qy, qz, qw  # quaternion
            else:
                self.stdout('Origin ERROR received')
                self.origin_error = True

    def update(self):
        packet = self.bus.listen()
        if packet is not None:
#            print('SubT', packet)
            timestamp, channel, data = packet
            if self.time is None or int(self.time.seconds)//60 != int(timestamp.seconds)//60:
                self.stdout(timestamp, '(%.1f %.1f %.1f)' % self.xyz if self.xyz is not None else '(xyz=None)', sorted(self.stat.items()))
                print(timestamp, list(('%.1f' % (v/100)) for v in self.voltage))
                self.stat.clear()

            self.time = timestamp

            if not self.is_virtual:
                self.sim_time_sec = self.time.total_seconds()

            self.stat[channel] += 1
            handler = getattr(self, "on_" + channel, None)
            if handler is not None:
                handler(timestamp, data)
            elif channel == 'scan' and not self.flipped:
                if self.last_send_time is not None and self.last_send_time - self.time > timedelta(seconds=0.1):
                    print('queue delay', self.last_send_time - self.time)
                self.scan = data
                if self.ref_scan is None or not equal_scans(self.scan, self.ref_scan, 200):
                    self.ref_scan = self.scan
                    self.ref_count = 0
                else:
                    self.ref_count += 1
                    if self.ref_count > 300:
                        print('Robot is stuck!', self.ref_count)
                        if self.collision_detector_enabled:
                            self.collision_detector_enabled = False
                            raise Collision()
                        self.ref_count = 0

                if self.local_planner is not None:
                    self.local_planner.update(data)
            elif channel == 'scan_back' and self.flipped:
                self.scan = data
                if self.local_planner is not None:
                    self.local_planner.update(data)
            elif channel == 'scan360':
                # reduce original 360 degrees scan to 270 degrees oriented forward or backward
                index45deg = int(round(len(data)/8))
                if self.flipped:
                    mid = len(data)//2
                    self.scan = (data[mid:]+data[:mid])[index45deg:-index45deg]
                else:
                    self.scan = data[index45deg:-index45deg]
                if self.local_planner is not None:
                    self.local_planner.update(data)
            elif channel == 'rot':
                temp_yaw, self.pitch, self.roll = [normalizeAnglePIPI(math.radians(x/100)) for x in data]
                if self.yaw_offset is None:
                    self.yaw_offset = -temp_yaw
                self.yaw = temp_yaw + self.yaw_offset

            elif channel == 'sim_time_sec':
                self.sim_time_sec = data
            elif channel == 'voltage':
                self.voltage = data
            elif channel == 'emergency_stop':
                self.emergency_stop = data

            elif channel == 'cmd':
                self.lora_cmd = data

            for m in self.monitors:
                m(self)
            return channel

    def wait(self, dt, use_sim_time=False):  # TODO refactor to some common class
        if use_sim_time:
            start_sim_time_sec = self.sim_time_sec
            while self.sim_time_sec - start_sim_time_sec < dt.total_seconds():
                self.update()
        else:
            if self.time is None:
                self.update()
            start_time = self.time
            while self.time - start_time < dt:
                self.update()

    def stdout(self, *args, **kwargs):
        output = StringIO()
        print(*args, file=output, **kwargs)
        contents = output.getvalue().strip()
        output.close()
        self.bus.publish('stdout', contents)
        print(contents)

#############################################
    def system_nav_trace(self, path=None):
        """
        Navigate along line
        """
        x0, y0, z0 = self.xyz
        trace = Trace()
        trace.add_line_to((x0, y0, z0))
        if path is not None:
            for x, y in path:
                trace.add_line_to((x - x0, y - y0, z0))
        trace.reverse()
        self.follow_trace(trace, timeout=timedelta(seconds=120), max_target_distance=2.5, safety_limit=0.2)

    def robust_follow_wall(self, radius, right_wall=False, timeout=timedelta(hours=3), dist_limit=None, flipped=False,
            pitch_limit=None, roll_limit=None):
        """
        Handle multiple re-tries with increasing distance from the wall if necessary
        """
        allow_virtual_flip = self.symmetric
        walldist = self.walldist
        total_dist = 0.0
        start_time = self.sim_time_sec
        overall_timeout = timeout
        while self.sim_time_sec - start_time < overall_timeout.total_seconds():
            if self.sim_time_sec - start_time > overall_timeout.total_seconds():
                print('Total Timeout Reached', overall_timeout.total_seconds())
                break
            timeout = timedelta(seconds=overall_timeout.total_seconds() - (self.sim_time_sec - start_time))
            print('Current timeout', timeout)

            dist, reason = self.follow_wall(radius=walldist, right_wall=right_wall, timeout=timeout, flipped=flipped,
                                    pitch_limit=self.limit_pitch, roll_limit=self.limit_roll)
            total_dist += dist
            if reason is None or reason in [REASON_LORA,]:
                break

            walldist += 0.2
            if not allow_virtual_flip:
                # Eduro not supported yet
                if reason in [REASON_VIRTUAL_BUMPER,]:
                    # virtual bumper already tried to backup a bit
                    continue
                # for large slope rather return home
                break

            # re-try with bigger distance
            print(self.time, "Re-try because of", reason)
            for repeat in range(2):
                self.follow_wall(radius=walldist, right_wall=not right_wall, timeout=timedelta(seconds=30), dist_limit=2.0,
                    flipped=not flipped, pitch_limit=self.return_limit_pitch, roll_limit=self.return_limit_roll)

                dist, reason = self.follow_wall(radius=walldist, right_wall=right_wall, timeout=timedelta(seconds=40), dist_limit=4.0,
                                        pitch_limit=self.limit_pitch, roll_limit=self.limit_roll, flipped=flipped)
                total_dist += dist
                if reason is None:
                    break
                if reason in [REASON_LORA, REASON_DIST_REACHED]:
                    break
                walldist += 0.2
            walldist = self.walldist
            if reason in [REASON_LORA,]:
                break


    def play_system_track(self):
        print("SubT Challenge Ver1!")
        try:
            with EmergencyStopMonitor(self):
                allow_virtual_flip = self.symmetric
                if distance(self.xyz, (0, 0)) > 0.1 or self.init_path is not None:
                    self.system_nav_trace(self.init_path)

#                self.go_straight(2.5)  # go to the tunnel entrance - commented our for testing
                walldist = self.walldist
                total_dist = 0.0
                start_time = self.sim_time_sec
                while self.sim_time_sec - start_time < self.timeout.total_seconds():
                    if self.sim_time_sec - start_time > self.timeout.total_seconds():
                        print('Total Timeout Reached', self.timeout.total_seconds())
                        break
                    timeout = timedelta(seconds=self.timeout.total_seconds() - (self.sim_time_sec - start_time))
                    print('Current timeout', timeout)

                    dist, reason = self.follow_wall(radius=walldist, right_wall=self.use_right_wall, timeout=timeout,
                                            pitch_limit=self.limit_pitch, roll_limit=self.limit_roll)
                    total_dist += dist
                    if reason is None or reason in [REASON_LORA,]:
                        break

                    walldist += 0.2
                    if not allow_virtual_flip:
                        # Eduro not supported yet
                        if reason in [REASON_VIRTUAL_BUMPER,]:
                            # virtual bumper already tried to backup a bit
                            continue
                        # for large slope rather return home
                        break

                    # re-try with bigger distance
                    print(self.time, "Re-try because of", reason)
                    for repeat in range(2):
                        self.follow_wall(radius=walldist, right_wall=not self.use_right_wall, timeout=timedelta(seconds=30), dist_limit=2.0,
                            flipped=allow_virtual_flip, pitch_limit=self.return_limit_pitch, roll_limit=self.return_limit_roll)

                        dist, reason = self.follow_wall(radius=walldist, right_wall=self.use_right_wall, timeout=timedelta(seconds=40), dist_limit=4.0,
                                                pitch_limit=self.limit_pitch, roll_limit=self.limit_roll)
                        total_dist += dist
                        if reason is None:
                            break
                        if reason in [REASON_LORA, REASON_DIST_REACHED]:
                            break
                        walldist += 0.2
                    walldist = self.walldist
                    if reason in [REASON_LORA,]:
                        break

                if self.use_return_trace and self.local_planner is not None:
                    self.stdout(self.time, "Going HOME %.3f" % dist, reason)
                    if allow_virtual_flip:
                        self.flipped = True  # return home backwards
                        self.scan = None
                    self.return_home(2 * self.timeout, home_threshold=1.0)
                    self.send_speed_cmd(0, 0)
                else:
                    print(self.time, "Going HOME", reason)
                    if not allow_virtual_flip:
                        self.turn(math.radians(90), timeout=timedelta(seconds=20))
                        self.turn(math.radians(90), timeout=timedelta(seconds=20))
                    self.robust_follow_wall(radius=self.walldist, right_wall=not self.use_right_wall, timeout=3*self.timeout, dist_limit=3*total_dist,
                            flipped=allow_virtual_flip, pitch_limit=self.return_limit_pitch, roll_limit=self.return_limit_roll)
                if self.artifacts:
                    self.bus.publish('artf_xyz', [[artifact_data, [round(x*1000), round(y*1000), round(z*1000)], self.robot_name, None]
                                              for artifact_data, (x, y, z) in self.artifacts])
        except EmergencyStopException:
            print(self.time, "EMERGENCY STOP - terminating")
        self.send_speed_cmd(0, 0)
        self.wait(timedelta(seconds=3))
#############################################

    def go_to_entrance(self):
        """
        Navigate to the base station tile end
        """
        trace = Trace()
        trace.update_trace(tuple(self.xyz))
        trace.add_line_to((-4.5, 0, self.height_above_ground))  # in front of the tunnel/entrance
        if self.use_right_wall:
            entrance_offset = -0.5
        elif self.use_center:
            entrance_offset = 0
        else:
            entrance_offset = 0.5
        trace.add_line_to((0.5, entrance_offset, self.height_above_ground))  # 0.5m inside, towards the desired wall.
        trace.reverse()
        self.follow_trace(trace, timeout=timedelta(seconds=30), max_target_distance=2.5, safety_limit=0.2)

    def play_virtual_part_enter(self):
        self.stdout("Waiting for origin ...")
        self.origin = None  # invalidate origin
        self.origin_error = False
        self.bus.publish('request_origin', True)
        while self.origin is None and not self.origin_error:
            self.update()
        self.stdout('Origin:', self.origin, self.robot_name)

        if self.origin is not None:
            x, y, z = self.origin
            heading = quaternion.heading(self.origin_quat)
            self.stdout('heading', math.degrees(heading), 'angle', math.degrees(math.atan2(-y, -x)), 'dist', math.hypot(x, y))

            self.go_to_entrance()
        else:
            # lost in tunnel
            self.stdout('Lost in tunnel:', self.origin_error)

    def play_virtual_part_explore(self):
        start_time = self.sim_time_sec
        for loop in range(100):
            self.collision_detector_enabled = True
            if self.sim_time_sec - start_time > self.timeout.total_seconds():
                print('Total Timeout Reached', self.timeout.total_seconds())
                break
            timeout = timedelta(seconds=self.timeout.total_seconds() - (self.sim_time_sec - start_time))
            print('Current timeout', timeout)

            dist, reason = self.follow_wall(radius=self.walldist, right_wall=self.use_right_wall,  # was radius=0.9
                                timeout=timeout, pitch_limit=self.limit_pitch, roll_limit=None)
            self.collision_detector_enabled = False
            if reason == REASON_VIRTUAL_BUMPER:
                assert self.virtual_bumper is not None
                self.virtual_bumper.reset_counters()

                # the robot ended in cycle probably
                self.return_home(timedelta(seconds=10))

                # try something crazy if you do not have other ideas ...
                before_center = self.use_center
                self.use_center = True
                dist, reason = self.follow_wall(radius=self.walldist, right_wall=self.use_right_wall,  # was radius=0.9
                                    timeout=timedelta(seconds=60), pitch_limit=self.limit_pitch, roll_limit=None)
                self.use_center = before_center
                if reason is None or reason != REASON_PITCH_LIMIT:
                    continue
            elif reason == REASON_LOOP:
                # Smaller walldist to reduce the chance that we miss the opening again that we missed before,
                # which made us end up in a loop.
                dist, reason = self.follow_wall(radius=self.walldist*0.75, right_wall=not self.use_right_wall,
                                    timeout=timedelta(seconds=5), pitch_limit=self.limit_pitch, roll_limit=None, detect_loop=False)
                if reason is None:
                    continue

            if reason is None or reason != REASON_PITCH_LIMIT:
                break
            self.stdout(self.time, "Microstep HOME %d %.3f" % (loop, dist), reason)
            self.go_straight(-0.3, timeout=timedelta(seconds=10))
            self.return_home(timedelta(seconds=10))

        self.stdout("Artifacts:", self.artifacts)

        self.stdout(self.time, "Explore phase finished %.3f" % dist, reason)

    def play_virtual_part_return(self, timeout):
        self.return_home(timeout)
        self.send_speed_cmd(0, 0)
        if self.artifacts:
            self.bus.publish('artf_xyz', [[artifact_data, [round(x * 1000), round(y * 1000), round(z * 1000)], self.robot_name, None]
                                          for artifact_data, (x, y, z) in self.artifacts])
        self.wait(timedelta(seconds=10), use_sim_time=True)
        self.stdout('Final xyz:', self.xyz)
        self.stdout('Final xyz (DARPA coord system):', self.xyz)

    def play_virtual_track(self):
        self.stdout("SubT Challenge Ver94!")
        self.stdout("Waiting for robot_name ...")
        while self.robot_name is None:
            self.update()
        self.stdout('robot_name:', self.robot_name)

        # wait for critical data
        while any_is_none(self.scan, self.yaw_offset, self.xyz):
            # self.xyz is initialized by pose2d or pose3d depending on robot type
            self.update()

        steps = parse_robot_name(self.robot_name)
        times_sec = [duration for action, duration in steps if action != 'home' and not action.startswith('enter')]
        self.stdout('Using times', times_sec)

        for action, duration, in steps:
            if action == 'wait':
                self.stdout(f'Wait for {duration} seconds')
                self.wait(timedelta(seconds=duration), use_sim_time=True)
                self.stdout('Artifacts collected during wait:', self.artifacts)

            elif action.startswith('enter'):
                self.use_right_wall = (action == 'enter-right')
                self.use_center = (action == 'enter-center')
                self.play_virtual_part_enter()

            elif action in ['left', 'right', 'center']:
                self.timeout = timedelta(seconds=duration)
                self.use_right_wall = (action == 'right')
                self.use_center = (action == 'center')
                self.play_virtual_part_explore()

            elif action == 'home':
                self.play_virtual_part_return(timedelta(seconds=duration))
            else:
                assert False, action  # unknown action

        self.wait(timedelta(seconds=30), use_sim_time=True)

#############################################

    def play(self):
        if self.is_virtual:
            return self.play_virtual_track()
        else:
            return self.play_system_track()

    def start(self):
        self.thread = threading.Thread(target=self.play)
        self.thread.start()

    def is_alive(self):
        return self.thread.is_alive()

    def request_stop(self):
        self.bus.shutdown()

    def join(self, timeout=None):
        self.thread.join(timeout)


def main():
    import argparse
    from osgar.lib.config import config_load
    #from osgar.record import record
    from osgar.zmqrouter import record

    parser = argparse.ArgumentParser(description='SubT Challenge')
    subparsers = parser.add_subparsers(help='sub-command help', dest='command')
    subparsers.required = True
    parser_run = subparsers.add_parser('run', help='run on real HW')
    parser_run.add_argument('config', nargs='+', help='configuration file')
    parser_run.add_argument('--note', help='add description')
    parser_run.add_argument('--walldist', help='distance for wall following (default: %(default)sm)', default=1.0, type=float)
    parser_run.add_argument('--side', help='which side to follow', choices=['left', 'right', 'auto'], required=True)
    parser_run.add_argument('--speed', help='maximum speed (default: from config)', type=float)
    parser_run.add_argument('--timeout', help='seconds of exploring before going home (default: %(default)s)',
                            type=int, default=10*60)
    parser_run.add_argument('--log', nargs='?', help='record log filename')
    parser_run.add_argument('--init-path', help='inital path to be followed from (0, 0). 2D coordinates are separated by ;')
    parser_run.add_argument('--start-paused', dest='start_paused', action='store_true',
                            help='start robota Paused and wait for LoRa Contine command')

    parser_replay = subparsers.add_parser('replay', help='replay from logfile')
    parser_replay.add_argument('logfile', help='recorded log file')
    parser_replay.add_argument('--force', '-F', dest='force', action='store_true', help='force replay even for failing output asserts')
    parser_replay.add_argument('--config', nargs='+', help='force alternative configuration file')
    args = parser.parse_args()

    if args.command == 'replay':
        from osgar.replay import replay
        args.module = 'app'
        app = replay(args, application=SubTChallenge)
        app.play()

    elif args.command == 'run':
        import logging
        logging.basicConfig(
            level=logging.DEBUG,
            format='%(asctime)s %(name)-16s %(levelname)-8s %(message)s',
        )
        # To reduce latency spikes as described in https://morepypy.blogspot.com/2019/01/pypy-for-low-latency-systems.html.
        # Increased latency leads to uncontrolled behavior and robot either missing turns or hitting walls.
        # Disabled garbage collection needs to be paired with gc.collect() at place(s) that are not time sensitive.
        gc.disable()

        cfg = config_load(*args.config, application=SubTChallenge)

        # apply overrides from command line
        cfg['robot']['modules']['app']['init']['walldist'] = args.walldist
        if args.side == 'auto':
            cfg['robot']['modules']['app']['init']['right_wall'] = 'auto'
        else:
            cfg['robot']['modules']['app']['init']['right_wall'] = args.side == 'right'
        cfg['robot']['modules']['app']['init']['timeout'] = args.timeout
        if args.init_path is not None:
            cfg['robot']['modules']['app']['init']['init_path'] = args.init_path

        if args.speed is not None:
            cfg['robot']['modules']['app']['init']['max_speed'] = args.speed

        cfg['robot']['modules']['app']['init']['start_paused'] = args.start_paused

        prefix = os.path.basename(args.config[0]).split('.')[0] + '-'
        record(cfg, prefix, args.log)


if __name__ == "__main__":
    main()

# vim: expandtab sw=4 ts=4



import unittest
import datetime
from unittest.mock import MagicMock

from osgar.drivers.lord_imu import get_packet, verify_checksum, parse_packet, LordIMU
from osgar.bus import Bus


SAMPLE_DATA = b'\x0e\x06=\xea\x01\x19>\xc3R\xd2>\xdb?V\x1e\x93ue\x80*\x0e\x04<\x87C\x97\xbdF%\xd4\xbf\x7f\xb4\x14\x0e\x05\xba.\xd6S\xba)b\xeb\xb9\xf6\xf7\x80\x0e\x06=\xeaD\x10>\xc3\\\xd5>\xdb4\x1d\xda\x04ue\x80*\x0e\x04<\x9d\x1c\xa3\xbdS\x19\xd9\xbf\x7f\xbc^\x0e\x05\xbas\xd7W\xb9\x15\x03\xa0\xbaD\xffc\x0e\x06=\xea'

SAMPLE_GPS_DATA = bytes.fromhex('756581432C034048F4AC14C660A2402C66410EBE08E44071A36872B020C5406DB35C28F5C2904008E5604062F1AA001F0F0807E30610090D11000000000003080B000900000007D716')


class LordIMUTest(unittest.TestCase):

    def test_parse_packet(self):
        self.assertEqual(get_packet(b''), (None, b''))

        buf = SAMPLE_DATA
        packet, buf = get_packet(buf)
        self.assertTrue(packet.startswith(b'\x75\x65'), packet)
        self.assertEqual(len(packet), 42 + 4 + 2)

        acc, gyro, mag, quat, gps = parse_packet(packet)

    def test_checksum(self):
        packet = bytes.fromhex("7565 0C20 0D08 0103 1200 0A04 000A 0500 0A13 0A01 0511 000A 1000 0A01 000A 0200 0A03 000A D43D")
        verify_checksum(packet)

        ping = bytes.fromhex('7565 0102 0201 E0C6')
        verify_checksum(ping)
        #ping 0xe4c6
        #replay: 7565 0104 04F1 0100 D56A
        # set to idle: 7565 0102 0202 E1C7
        # ping         7565 0102 0201 E0C6

    def test_node(self):
        logger = MagicMock()
        bus = Bus(logger)
        tester = bus.handle('tester')
        tester.register('raw')
        imu = LordIMU(config={}, bus=bus.handle('lord'))
        bus.connect('tester.raw', 'lord.raw')
        imu.start()
        tester.publish('raw', SAMPLE_DATA)
        imu.request_stop()
        imu.join()
        self.assertEqual(imu.raw, SAMPLE_DATA)
#        q.put.assert_called_once_with((135, 'raw',
#            b'\x00\x00\x03\x01\x01\xfb'))

    def test_parse_gps(self):
        verify_checksum(SAMPLE_GPS_DATA)
        __, __, __, __, gps = parse_packet(SAMPLE_GPS_DATA)
        self.assertEqual(gps, ((49.9115015, 14.199715099999999), 9,
                               datetime.datetime(2019, 6, 16, 9, 13, 17)))

    def test_index_error(self):
        """
           bug - used to raise Index Error:
File "M:\git\osgar\osgar\drivers\lord_imu.py", line 28, in get_packet
    packet_size = buf[i + 3] + 4 + 2  # + header size + checksum
IndexError: index out of range
"""
        packet, buf = get_packet(b'ue')
        self.assertIsNone(packet)
        self.assertEqual(buf, b'ue')

    def test_invalid_checksum(self):
        raw_data = bytes.fromhex('75658252100d3d3302803c68b2e03cee54000001100e3bb32f1f3b0ed8403a637b3e00010e14000000000'
                                 '00000000001000214033c7a802ebf4e2607bf17b2a78ae04161ef4e0001100ebd08efc83ebe409a3cb8df380001')
        self.assertFalse(verify_checksum(raw_data))

        packet, new_buf = get_packet(raw_data + bytes.fromhex('1234'))
        self.assertIsNone(packet)
        self.assertEqual(new_buf, bytes.fromhex('1234'))

# vim: expandtab sw=4 ts=4



import unittest
import math

from .line import Line, pointAtPolyLineDist


class LineTest(unittest.TestCase):

    def test_signed_distance(self):
        line = Line((0, 0), (1, 1))
        self.assertAlmostEqual(line.signedDistance((0.5, 0.5)), 0.0)
        self.assertAlmostEqual(line.signedDistance((0, 1)), math.sqrt(1/2))
        self.assertAlmostEqual(line.signedDistance((1, 0)), -math.sqrt(1/2))

        # it is a infinite line
        self.assertAlmostEqual(line.signedDistance((2, 2)), 0.0)

    def test_distance_to_finish(self):
        line = Line((0, 0), (1, 1))
        self.assertAlmostEqual(line.distanceToFinishLine((0, 0)), math.sqrt(2))
        self.assertAlmostEqual(line.distanceToFinishLine((-1, 1)), math.sqrt(2))
        self.assertAlmostEqual(line.distanceToFinishLine((2, 2)), -math.sqrt(2))

        self.assertTrue(line.finished((2, 2)))

    def test_snap(self):
        line = Line((0, 0), (1, 1))
        p = line.snap((1, 0))
        self.assertAlmostEqual(p[0], 0.5)
        self.assertAlmostEqual(p[1], 0.5)

        p = line.snap((2, 2))
        self.assertAlmostEqual(p[0], 2)
        self.assertAlmostEqual(p[1], 2)

    def test_nearest(self):
        line = Line((0, 0), (1, 1))
        p, abs_dist, index = line.nearest((2, 2))
        self.assertAlmostEqual(p[0], 1)
        self.assertAlmostEqual(p[1], 1)
        self.assertAlmostEqual(abs_dist, math.sqrt(2))
        self.assertEqual(index, 1)

        p, abs_dist, index = line.nearest((1, 0))
        self.assertAlmostEqual(p[0], 0.5)
        self.assertAlmostEqual(p[1], 0.5)
        self.assertAlmostEqual(abs_dist, abs(-math.sqrt(1/2)))
        self.assertEqual(index, -1)

    def test_point_at_dist(self):
        polyline = [(0, 0), (1, 0), (1, 2)]
        p = pointAtPolyLineDist(polyline, 2.0)
        self.assertAlmostEqual(p[0], 1)
        self.assertAlmostEqual(p[1], 1)

# vim: expandtab sw=4 ts=4




"""
  OSGAR ArtifactDetectorDNN wrapper for DNN detector
"""
import os.path
from io import StringIO

import cv2
import numpy as np
from subt.tf_detector import CvDetector
try:
    import torch
    import subt.artf_model
    from subt.artf_detector import Detector
except ImportError:
    print('\nWarning: missing torch!\n')

from osgar.node import Node
from osgar.bus import BusShutdownException
from subt.artifacts import (RESCUE_RANDY, BACKPACK, PHONE, HELMET, ROPE)


NAME2IGN = {
    'survivor': RESCUE_RANDY,
    'backpack': BACKPACK,
    'phone': PHONE,
    'helmet': HELMET,
    'rope': ROPE
}


def check_results(result_mdnet, result_cv):
    result = result_mdnet.copy()
    ret = []
    if len(result) == 0 or len(result_cv) == 0:
        return ret
    for r_cv in result_cv:
        name_cv, score_cv, bbox = r_cv
        bbox_x1, bbox_y1, bbox_x2, bbox_y2 = bbox
        ret_points = []
        for r in result.copy():
            name, points = r
            if name != name_cv:
                continue
            x = np.array([p[0] for p in points])
            y = np.array([p[1] for p in points])
            x_in_bbox = np.logical_and((x > bbox_x1), (x < bbox_x2))
            y_in_bbox = np.logical_and((y > bbox_y1), (y < bbox_y2))
            if np.any(np.logical_and(x_in_bbox, y_in_bbox)):  # at least one point is in the bbox
                ret_points.extend(points)
                result.remove(r)
        if ret_points:
            ret.append((name_cv, ret_points))
    return ret


def result2report(result, depth, fx):
    """return relative XYZ distances to camera"""
    if depth is None:
        return None  # ignore detected artifacts for missing depth data
                     # typically some glitch on start
    width = depth.shape[1]
    height = depth.shape[0]
    x_arr = [x for x, y, certainty in result[0][1]]  # ignore multiple objects
    y_arr = [y for x, y, certainty in result[0][1]]  # ignore multiple objects
    dist = [depth[y][x] for x, y, certainty in result[0][1]]  # ignore multiple objects
    if 0xFFFF in dist:
        return None  # out of range
    x_min, x_max = min(x_arr), max(x_arr)
    y_min, y_max = min(y_arr), max(y_arr)
    scale = np.median(dist)
    rel_x = int(scale)  # relative X-coordinate in front
    rel_y = int(scale * (width/2 - (x_min + x_max)/2)/fx)  # Y-coordinate is to the left
    rel_z = int(scale * (height/2 - (y_min + y_max)/2)/fx)  # Z-up
    return [NAME2IGN[result[0][0]], [rel_x, rel_y, rel_z]]


class ArtifactDetectorDNN(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register("artf", "dropped", "debug_artf", "debug_depth:gz", "stdout",
                     "debug_result", "debug_cv_result")
        self.time = None
        self.width = None  # not sure if we will need it
        self.depth = None  # more precise artifact position from depth image
        self.cv_detector = CvDetector().subt_detector
        self.detector = self.create_detector()
        self.fx = config.get('fx', 554.25469)  # use drone X4 for testing (TODO update all configs!)

    def create_detector(self):
        model = os.path.join(os.path.dirname(__file__), '../../../mdnet3.64.64.13.4.relu.pth')
        confidence_thresholds = {
            'survivor': 0.8,
            'backpack': 0.8,
            'phone': 0.5,
            'helmet': 0.5,
            'rope': 0.6
        }
        max_gap = 16
        min_group_size = 2

        use_cuda = torch.cuda.is_available()
        device = torch.device("cuda" if use_cuda else "cpu")
        print('Using:', device)
        model, categories = subt.artf_model.load_model(model, device)
        return Detector(model, confidence_thresholds, categories, device,
                        max_gap, min_group_size)

    def wait_for_image(self):
        channel = ""
        while channel != "image":
            self.time, channel, data = self.listen()
            setattr(self, channel, data)
        return self.time

    def stdout(self, *args, **kwargs):
        # maybe refactor to Node?
        output = StringIO()
        print(*args, file=output, **kwargs)
        contents = output.getvalue().strip()
        output.close()
        self.publish('stdout', contents)
        print(self.time, contents)

    def run(self):
        try:
            dropped = 0
            while True:
                now = self.publish("dropped", dropped)
                dropped = -1
                timestamp = now
                while timestamp <= now:
                    timestamp = self.wait_for_image()
                    dropped += 1
                self.detect(self.image)
        except BusShutdownException:
            pass

    def detect(self, image):
        img = cv2.imdecode(np.fromstring(image, dtype=np.uint8), 1)
        if self.width is None:
            self.stdout('Image resolution', img.shape)
            self.width = img.shape[1]
        assert self.width == img.shape[1], (self.width, img.shape[1])

        result = self.detector(img)
        result_cv = self.cv_detector(img)
        if result or result_cv:
            # publish the results independent to detection validity
            self.publish('debug_result', result)
            self.publish('debug_cv_result', result_cv)
            checked_result = check_results(result, result_cv)
            if checked_result:
                report = result2report(checked_result, self.depth, self.fx)
                if report is not None:
                    self.publish('artf', report)
                    self.publish('debug_artf', image)  # JPEG
                    self.publish('debug_depth', self.depth)


if __name__ == "__main__":
    # run "replay" without calling detections - only XYZ offset check
    import argparse
    from datetime import timedelta

    from osgar.lib.serialize import deserialize
    from osgar.logger import LogReader, lookup_stream_id, lookup_stream_names
    from ast import literal_eval

    parser = argparse.ArgumentParser(description='Test 3D reports')
    parser.add_argument('logfile', help='OSGAR logfile')
    parser.add_argument('--time-limit-sec', '-t', help='cut time in seconds', type=float)
    parser.add_argument('--verbose', '-v', help="verbose mode", action='store_true')
    args = parser.parse_args()

    names = lookup_stream_names(args.logfile)
    assert 'detector.artf' in names, names  # XYZ relative offsets
    assert 'detector.debug_depth' in names, names
    assert 'detector.debug_result' in names, names
    assert 'detector.debug_cv_result' in names, names

    artf_stream_id = names.index('detector.artf') + 1
    depth_stream_id = names.index('detector.debug_depth') + 1
    result_id = names.index('detector.debug_result') + 1
    cv_result_id = names.index('detector.debug_cv_result') + 1

    # read config file from log
    with LogReader(args.logfile, only_stream_id=0) as log:
        print("original args:", next(log)[-1])  # old arguments
        config_str = next(log)[-1]
        config = literal_eval(config_str.decode('ascii'))

    assert 'detector' in config['robot']['modules']
    fx = config['robot']['modules']['detector']['init']['fx']

    depth = None
    last_artf = None  # reported before debug_depth
    last_result = None
    last_cv_result = None
    with LogReader(args.logfile,
                   only_stream_id=[artf_stream_id, depth_stream_id, result_id, cv_result_id]) as logreader:
        for time, stream, msg_data in logreader:
            if args.time_limit_sec is not None and time.total_seconds() > args.time_limit_sec:
                break
            data = deserialize(msg_data)

            if stream == depth_stream_id:
                depth = data
                # debug_depth is stored ONLY when both detector detect something and it is fused
                assert last_result is not None
                assert last_cv_result is not None

                checked_result = check_results(last_result, last_cv_result)
                assert checked_result  # the debug depth is stored, so there should valid report
                report = result2report(checked_result, depth, fx)
                if args.verbose:
                    print(report)
                assert last_artf == report, (last_artf, report)

            elif stream in [result_id, cv_result_id]:
                if args.verbose:
                    print(time, data)
                if stream == result_id:
                    last_result = data
                elif stream == cv_result_id:
                    last_cv_result = data
                else:
                    assert False, stream

            elif stream == artf_stream_id:
                if args.verbose:
                    print(time, 'Original report:', data)
                last_artf = data
                assert last_artf is not None, time
            else:
                assert False, stream  # unexpected stream

# vim: expandtab sw=4 ts=4



""""
  Extract robot positions from Ignition SQLite log file
"""
import sqlite3
import struct

g_robots = {}

def unpack_protobuf(data, depth=0, prefix=None, verbose=False):
    robot_name, robot_xyz = None, None
    if depth >= 3:
        return
    if prefix is None:
        prefix = []
    if verbose:
        print(prefix, data[:100].hex())
    pos = 0
    while pos < len(data):
        c = data[pos]
        field_number = c >> 3
        wire_type = c & 0x7
        if verbose:
            print(pos, field_number, wire_type)
        pos += 1
        if wire_type == 0:  # variant
            start = pos
            while data[pos] & 0x80 == 0x80:
                pos += 1
            pos += 1
#            print('Var', field_number, data[start:pos].hex())
        elif wire_type == 1:  # 64bit
            d = struct.unpack('<d', data[pos:pos+8])[0]
#            i = struct.unpack('<q', data[:8])[0]
            if prefix == [2, 4]:  # xyz
#                print('double', prefix, field_number, d)
                if field_number == 2:
                    x = d
                elif field_number == 3:
                    y = d
                elif field_number == 4:
                    z = d
                    robot_xyz = (x, y, z)
                    return robot_xyz
#            print('int', prefix, field_number, i)
            pos += 8
        elif wire_type == 2:
            size = data[pos]
            pos += 1

            name = data[pos:pos + size]
            if verbose:
                print(name)
            if name in [b'base_link', b'front_left_wheel', b'front_right_wheel', b'rear_left_wheel', b'rear_right_wheel']:
#                print('terminated', name)
                return
            if prefix == [2] and field_number == 2:
#                print(prefix, field_number, wire_type, size, data[pos:pos + size])
                robot_name = data[pos:pos + size].decode('ascii')
            key = prefix + [field_number]
            if field_number in [2, 4, 5]:
#                print('Key', key)
                if key != [2, 2]:
                    ret = unpack_protobuf(data[pos:pos + size], depth=depth+1, prefix=key)
                    if ret is not None:
#                        print('return', ret)
                        if len(ret) == 3:
                            robot_xyz = ret
                pos += size
        else:
            assert False, wire_type
    if robot_name is not None:
        global g_robots
        g_robots[robot_name] = robot_xyz
#        print(robot_name, robot_xyz)
        return robot_name, robot_xyz
    return None


def _extract_poses_gen(filename):
    con = sqlite3.connect(filename)
    cursor = con.cursor()

    # overview of available tables
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    tables = cursor.fetchall()
    for table in tables:
        cursor.execute("SELECT sql FROM sqlite_master WHERE type='table' and name='%s';" % table)
#        print(cursor.fetchall()[0][0])
#        print()

    # start time?
    cursor.execute("SELECT * FROM migrations;")
    print('Start record:', cursor.fetchall())

    cursor.execute("SELECT * FROM topics;")
    topics = cursor.fetchall()
    print(topics)

    cursor.execute(r"SELECT id FROM topics where name LIKE '%/dynamic_pose/info';")
    result = cursor.fetchall()
    dynamic_topic_id = result[0][0]

    cursor.execute("SELECT * FROM messages;")
    for i, row in enumerate(cursor):
        index, time_ns, topic_id, data = row
        if topic_id == dynamic_topic_id:  # '/world/urban_circuit_practice_03/dynamic_pose/info'
#            print(data[:100].hex())
            unpack_protobuf(data)
            if len(g_robots) > 0:
                yield g_robots
#            print(index, data, len(data))
#           print(index, len(data))
#        if b'A60F900L' in data:  #i > 100:
#        if b'E180F200F700L' in data:  #i > 100:
#            assert b'\x08A60F900L' in data, data
#            print(data[:30])
#            print(data)
#            break
            """
            print('----------------')
            index = 0
#            while index < len(data):
            for j in range(10):
                size = data[index]
                index += 1
                print(size, data[index:index + size])
                index += size
                print(index, len(data))
                """
#        if i > 100:
#            break


def extract_poses_gen(filename):
    try:
        for data in _extract_poses_gen(filename):
            yield  data
    except sqlite3.DatabaseError as e:
        print(e)


if __name__ == "__main__":
    import argparse
    from subt.log2map import pts2image

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('ignfile', help='Ignition file from simulation (state.tlog)')
    parser.add_argument('--num', '-n', help='number of records', type=int)
    parser.add_argument('--only-robot', help='name prefix of selected robot')
    parser.add_argument('-v', '--verbose', help='verbose mode', action='store_true')
    parser.add_argument('--out', help='optional image output (filename)')
    args = parser.parse_args()

    pts = []
    for i, robots in enumerate(extract_poses_gen(args.ignfile)):
        if args.verbose:
            print(i, robots)
        for r in robots.items():
            name, xyz = r
            x, y, z = xyz
            if args.only_robot is  None or name.startswith(args.only_robot):
                pts.append((int(x * 1000), int(y * 1000)))  # TODO change to meters
        if args.num is not None and i >= args.num:
            break
    print('Total:', i)

    if args.out is not None:
        pts2image(pts, [], args.out)

# vim: expandtab sw=4 ts=4



import unittest
from pathlib import Path

import cv2
import numpy as np

import subt.visual_odometry as vo

curdir = Path(__file__).parent

def odometry_params_for_tests():
    return vo.VisualOdometryParams(
            camera_focal_length=462.1, camera_center=(320.5, 180.5))


def load_test_data():
    bundle = np.load(str(curdir/'test_data/rgbd.npz'))
    return bundle['color_img'], bundle['grey_img'], bundle['depth']


def odom_estimate(odom, observations_sequence):
    for observations in observations_sequence:
        img = observations['img']
        depth = observations.get('depth', np.zeros(img.shape, dtype=np.float32))
        rpy = observations.get('rpy', None)
        rot = None if rpy is None else vo.euler_angles_to_rotation_matrix(*rpy)
        altitude = observations.get('altitude', None)
        odom.update(img, depth, rot, altitude)
    return odom.xyz, odom.rpy


class VisualOdometryTest(unittest.TestCase):
    def test_as_rigid_transform_xyz(self):
        transform = vo.as_rigid_transform(xyz=[1, 2, 3])
        pt = np.asarray([2., 2., 2., 1.])
        transformed_pt = (transform @ pt)[:3]
        self.assertTrue(np.all(transformed_pt == [3, 4, 5]))

    def test_as_rigid_transform_rpy(self):
        transform = vo.as_rigid_transform(rpy=np.radians((0, 0, 90)))
        pt = np.asarray([1., 2., 3., 1.])
        transformed_pt = (transform @ pt)[:3]
        self.assertTrue(np.all(np.isclose(transformed_pt, [-2, 1, 3])))

    def test_as_rigid_transform_rot(self):
        transform = vo.as_rigid_transform(rot=
                vo.euler_angles_to_rotation_matrix(*np.radians((0, 0, 90))))
        pt = np.asarray([1., 2., 3., 1.])
        transformed_pt = (transform @ pt)[:3]
        self.assertTrue(np.all(np.isclose(transformed_pt, [-2, 1, 3])))

    def test_inverse_rotation(self):
        # Some not very specific rotation.
        rot = vo.euler_angles_to_rotation_matrix(0.13, 0.24, 0.42)
        inv_rot = vo.inverse_rotation(rot)
        self.assertTrue(np.all(np.isclose(inv_rot @ rot, np.eye(3))))

    def test_inverse_rigid_transform(self):
        rt = vo.as_rigid_transform(xyz=[1, 2, 3], rpy=[0.13, 0.24, 0.42])
        irt = vo.inverse_rigid_transform(rt)
        self.assertTrue(np.all(np.isclose(irt @ rt, np.eye(4))))

    def test_euler_to_rot_and_back(self):
        orig_rpy = (-np.pi / 3, np.pi / 4., np.pi / 2)
        rot = vo.euler_angles_to_rotation_matrix(*orig_rpy)
        back_rpy = vo.rotation_matrix_to_euler_angles(rot)
        self.assertEqual(back_rpy, orig_rpy)

    def test_near_rotation_matrix(self):
        orig_rot = np.eye(3)
        messed_up_rot = orig_rot * (1 + 1e-3)
        messed_up_rot[0, 2] = 1e-12
        recovered_rot = vo.near_rotation_matrix(messed_up_rot)

        # The disturbance we caused is noticeable ...
        self.assertFalse(np.all(np.isclose(messed_up_rot, orig_rot)))
        # ... and we are able to recover from it.
        self.assertTrue(np.all(np.isclose(recovered_rot, orig_rot)))

    def test_camera_odometry_no_move(self):
        color_img, grey_img, depth = load_test_data()
        h, w = grey_img.shape
        odom = vo.CameraVisualOdometry(odometry_params_for_tests())
        # If the scene hasn't changed, the estimate should say that
        # the camera has not moved.
        xyz, rpy = odom_estimate(
                odom,
                [{'img' : grey_img, 'depth' : depth},
                 {'img' : grey_img, 'depth' : depth}])
        self.assertEqual(xyz.tolist(), [0, 0, 0])
        self.assertEqual(rpy, (0, 0, 0))

    def test_camera_odometry_no_move_less_obvious(self):
        color_img, grey_img, depth = load_test_data()
        h, w = grey_img.shape
        odom = vo.CameraVisualOdometry(odometry_params_for_tests())

        # Next scene, slightly modified so that it is not obvious that it is
        # the same one.
        next_grey_img = grey_img.copy()
        next_grey_img[:64,:] = 0
        next_depth = depth.copy()
        next_depth[:64,:] = 0

        # If the scene hasn't changed, the estimate should say that
        # the camera has not moved.
        xyz, rpy = odom_estimate(
                odom,
                [{'img' : grey_img, 'depth' : depth},
                 {'img' : next_grey_img, 'depth' : next_depth}])
        self.assertLess(np.linalg.norm(xyz), 1e-3)
        self.assertLess(np.linalg.norm(rpy), np.radians(0.01))

    def test_camera_odometry_forward(self):
        color_img, grey_img, depth = load_test_data()
        h, w = grey_img.shape

        # Scaling the image up roughly corresponds to moving directly forward
        # from the point of view of the camera. We need to keep the 16:9 image
        # ratio of the test input during the resize.
        self.assertEqual(grey_img.shape, (360, 640))
        fw_img = cv2.resize(grey_img[9:-9,16:-16], (w, h))
        self.assertEqual(grey_img.shape, fw_img.shape)

        odom = vo.CameraVisualOdometry(odometry_params_for_tests())
        xyz, rpy = odom_estimate(
                odom, [{'img' : grey_img, 'depth' : depth}, {'img' : fw_img}])
        # The estimate should lead to a substantial move in the direction of x
        # axis, i.e. bigger move in that direction than other directions.
        # There also shouldn't be a big change in heading.
        self.assertTrue(xyz[0] > 0)
        A_LOT = 4
        self.assertGreater(xyz[0], A_LOT * abs(xyz[1]))
        self.assertGreater(xyz[0], A_LOT * abs(xyz[2]))
        NOT_MUCH = np.radians(2)
        self.assertTrue(np.all(np.abs(rpy) < NOT_MUCH))

    def test_camera_odometry_rot_left(self):
        color_img, grey_img, depth = load_test_data()
        h, w = grey_img.shape

        # Shifting the image to the right mimics rotation to the left on the
        # spot.
        rot_img = np.zeros_like(grey_img)
        rot_img[:,16:] = grey_img[:,:-16]
        odom = vo.CameraVisualOdometry(odometry_params_for_tests())
        xyz, rpy = odom_estimate(
                odom, [{'img' : grey_img, 'depth' : depth}, {'img' : rot_img}])
        # The expected translation should be small.
        NOT_MUCH = 0.05
        self.assertLess(np.linalg.norm(xyz), NOT_MUCH)
        # The rotation should be bigger than the other rotations.
        A_LOT = 10
        self.assertGreater(rpy[2], A_LOT * rpy[0])
        self.assertGreater(rpy[2], A_LOT * rpy[1])

    def test_camera_odometry_imu_correction(self):
        color_img, grey_img, depth = load_test_data()
        h, w = grey_img.shape

        # Scaling the image up roughly corresponds to moving directly forward
        # from the point of view of the camera. We need to keep the 16:9 image
        # ratio of the test input during the resize.
        self.assertEqual(grey_img.shape, (360, 640))
        fw_img = cv2.resize(grey_img[9:-9,16:-16], (w, h))

        odom = vo.CameraVisualOdometry(odometry_params_for_tests())
        xyz, rpy = odom_estimate(
                odom,
                [{'img' : grey_img, 'depth' : depth},
                 {'img' : fw_img, 'rpy' : (0, 0, 0)}])
        # The estimate should lead to a substantial move in the direction of x
        # axis, i.e. bigger move in that direction than other directions.
        # There also shouldn't be almost any change in heading, because we
        # provided information about heading.
        self.assertTrue(xyz[0] > 0)
        A_LOT = 4
        self.assertGreater(xyz[0], A_LOT * abs(xyz[1]))
        self.assertGreater(xyz[0], A_LOT * abs(xyz[2]))
        NOT_MUCH = np.radians(1e-6)
        self.assertTrue(np.all(np.abs(rpy) < NOT_MUCH))

    def test_camera_odometry_altitude_correction(self):
        color_img, grey_img, depth = load_test_data()
        h, w = grey_img.shape
        # Pretending forward motion.
        fw_img = cv2.resize(grey_img[9:-9,16:-16], (w, h))
        odom = vo.CameraVisualOdometry(odometry_params_for_tests())
        self.assertEqual(odom.xyz[2], 0.)
        xyz, rpy = odom_estimate(
                odom,
                [{'img' : grey_img, 'depth' : depth},
                 {'img' : fw_img, 'altitude' : 0.25}])
        # Despite odometry starting at (0, 0, 0) and estimating pure forward
        # motion, the altitude mesurement should pust the estimate up.
        self.assertEqual(xyz[2], 0.25)


    def test_robot_odometry_forward(self):
        color_img, grey_img, depth = load_test_data()
        h, w = grey_img.shape

        # Scaling the image up roughly corresponds to moving directly forward
        # from the point of view of the camera. We need to keep the 16:9 image
        # ratio of the test input during the resize.
        self.assertEqual(grey_img.shape, (360, 640))
        fw_img = cv2.resize(grey_img[9:-9,16:-16], (w, h))

        odom = vo.RobotVisualOdometry(
                camera_xyz=(0.2, 0.2, 0.2),
                # Camera points to the left.
                camera_rpy=(0, 0, np.pi/2),
                odometry_params=odometry_params_for_tests())
        xyz, rpy = odom_estimate(
                odom, [{'img' : grey_img, 'depth' : depth}, {'img' : fw_img}])

        # The estimate should lead to a substantial move in the direction of y
        # axis, because camera looks to the left and the move was forward from
        # its perspective. I.e. the should be a bigger move in y direction
        # than other directions.
        # There also shouldn't be a big change in heading.
        self.assertTrue(xyz[1] > 0)
        A_LOT = 4
        self.assertGreater(xyz[1], A_LOT * abs(xyz[0]))
        self.assertGreater(xyz[1], A_LOT * abs(xyz[2]))
        NOT_MUCH = np.radians(2)
        self.assertTrue(np.all(np.abs(rpy) < NOT_MUCH))

    def test_global_odometry_forward(self):
        color_img, grey_img, depth = load_test_data()
        h, w = grey_img.shape

        # Scaling the image up roughly corresponds to moving directly forward
        # from the point of view of the camera. We need to keep the 16:9 image
        # ratio of the test input during the resize.
        self.assertEqual(grey_img.shape, (360, 640))
        fw_img = cv2.resize(grey_img[9:-9,16:-16], (w, h))

        odom = vo.GlobalVisualOdometry(
                initial_xyz=(0, 0, 0),
                # Robot starts looking "backwards".
                initial_rpy=(0, 0, np.pi),
                camera_xyz=(0.2, 0.2, 0.2),
                # Camera points to the left of the robot.
                camera_rpy=(0, 0, np.pi/2),
                odometry_params=odometry_params_for_tests())
        xyz, rpy = odom_estimate(
                odom, [{'img' : grey_img, 'depth' : depth}, {'img' : fw_img}])

        # The estimate should lead to a substantial move in the direction of
        # negative y axis, because camera looks to the left of the robot and
        # robot starts at a position looking in the direction of negative X
        # axis and the move was forward from the perspective of the camera.
        # I.e. the should be a bigger move in -y direction than other
        # directions.
        # There also shouldn't be a big change in heading.
        self.assertTrue(xyz[1] < 0)
        A_LOT = 4
        self.assertGreater(-xyz[1], A_LOT * abs(xyz[0]))
        self.assertGreater(-xyz[1], A_LOT * abs(xyz[2]))
        NOT_MUCH = np.radians(2)
        expected_rpy = np.array([0, 0, np.pi])
        self.assertTrue(np.all(np.abs(rpy - expected_rpy) < NOT_MUCH))


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import MagicMock

from osgar.drivers.canserial import CANSerial, CAN_packet


class CANSerialTest(unittest.TestCase):

    def test_split_buffer(self):
        self.assertEqual(CANSerial.split_buffer(b''), (b'', b''))
        self.assertEqual(CANSerial.split_buffer(b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xfe\x10'), (b'', b'\xfe\x10'))
        data = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xfeW\xfe0@h\x9e\x01i\x01\xf7\x01\x18\x00'
        self.assertEqual(CANSerial.split_buffer(data), (b'\xfe0@h\x9e\x01i\x01\xf7\x01\x18\x00', b'\xfeW'))

        data = b'0@h\x9e\x01i\x01\xf7\x01\x18\x00'
        self.assertEqual(CANSerial.split_buffer(data), (b'h\x9e\x01i\x01\xf7\x01\x18\x00', b'0@'))

    def test_can_packet(self):
        self.assertEqual(CAN_packet(0x400, [0, 0]), b'\x80\x02\x00\x00')

    def test_can_bridge_initialization(self):
        bus = MagicMock()
        can = CANSerial(config={}, bus=bus)
        self.assertFalse(can.can_bridge_initialized)

        for __ in can.process_gen(b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xfe\x10'):
            pass
        self.assertTrue(can.can_bridge_initialized)

        bus.publish.assert_called_with('raw', b'\xfe1')

    def test_invalid_config(self):
        bus = MagicMock()
        config =  {"speed": 123, "canopen":True}
        with self.assertRaises(ValueError) as err:
            can = CANSerial(config=config, bus=bus)
        self.assertEqual(str(err.exception), "unsupported speed: 123\n" +
            "Use:['10k', '20k', '50k', '125k', '250k', '500k', '800k', '1M']")

    def test_reset_module(self):
        bus = MagicMock()
        can = CANSerial(config={}, bus=bus)

        self.assertEqual(can.modules_for_restart, set())
        can.check_and_restart_module(2, 127)
        bus.publish.assert_called_with('raw', b'\x00\x02\x81\x02')
        self.assertEqual(can.modules_for_restart, set([2]))

        can.check_and_restart_module(2, 127)
        bus.publish.assert_called_with('raw', b'\x00\x02\x01\x02')

        can.check_and_restart_module(2, 5)
        self.assertEqual(can.modules_for_restart, set())

    def test_can_packet_bytes(self):
        # it should support both list of integers as well as payload bytes()
        self.assertEqual(CAN_packet(0x400, bytes([0, 0])), b'\x80\x02\x00\x00')

# vim: expandtab sw=4 ts=4



"""
    Geometry 2D line helper
"""

import math


def distance(planar1, planar2):
    """Distance of two planar points"""
    x = planar1[0] - planar2[0]
    y = planar1[1] - planar2[1]
    return math.hypot(x, y)


def pointAtPolyLineDist(pts, dist):
    """Return point on polyline (planar) in given distance from start"""
    if len(pts) == 0:
        return None
    if len(pts) == 1:
        return pts[0]

    prev = pts[0]
    distFromStart = 0
    for p in pts[1:]:
        distFromStart += distance( prev, p )
        if dist < distFromStart:
            break
        prev = p
    if distance( prev, p ) < 0.000001:
        return p
    # resize (p - prev) to dist-distFromStart
    assert dist < distFromStart, (dist, distFromStart)
    frac = (distFromStart - dist)/distance(prev, p)
    return p[0] + frac*(prev[0]-p[0]), p[1] + frac*(prev[1]-p[1])



class Line:
    """
       2D line with basic distance and snap operations
    """

    def __init__(self, start, end):
        """Define line with two points"""
        self.start = start
        self.end = end
        self.angle = math.atan2( end[1] - start[1], end[0] - start[0] )
        self._a = -math.sin( self.angle )
        self._b = math.cos( self.angle )
        self._c = -self._a * start[0] - self._b * start[1]
        self._perpC = self._b * end[0] - self._a * end[1] # perpendicular line in end position
        self.length = distance(start, end)

    def signedDistance(self, pos):
        """Distance from line with orientation - positive on left side"""
        return self._a * pos[0] + self._b * pos[1] + self._c

    def distanceToFinishLine(self, pos):
        """Distance to the end of line along the line"""
        # returns negative value when end was already crossed
        return - self._b * pos[0] + self._a * pos[1] + self._perpC;

    def finished(self, pos):
        return self.distanceToFinishLine(pos) <= 0

    def snap(self, pos):
        """Return nearest position on (infinite) line"""
        dist = self.distanceToFinishLine( pos )
        return self.end[0] - self._b * dist, self.end[1] + self._a * dist

    def nearest( self, pos ):
        """return abs distance and type 0=start, 1=end, -1=inside of line"""
        # the same as snap, but restricted to finite line
        dist = self.distanceToFinishLine(pos)
        if dist <= 0:
            return self.end, distance(pos, self.end), 1
        if dist >= self.length:
            return self.start, distance(pos, self.start), 0
        return self.snap(pos), math.fabs(self.signedDistance(pos)), -1

# vim: expandtab sw=4 ts=4



"""
  pose2d fusion
"""
import math

from osgar.node import Node


class FusionPose2d(Node):
    def __init__(self, config, bus):
        super().__init__(config, bus)
        bus.register('pose2d')
        self.yaw, self.pitch, self.roll = 0, 0, 0
        self.xyz = 0, 0, 0
        self.last_position = None

    def update(self):
        channel = super().update()
        if channel == 'pose2d':
            x, y, heading = self.pose2d
            pose = (x/1000.0, y/1000.0, math.radians(heading/100.0))
            if self.last_position is not None:
                dist = math.hypot(pose[0] - self.last_position[0], pose[1] - self.last_position[1])
                direction = ((pose[0] - self.last_position[0]) * math.cos(self.last_position[2]) +
                             (pose[1] - self.last_position[1]) * math.sin(self.last_position[2]))
                if direction < 0:
                    dist = -dist
            else:
                dist = 0.0
            self.last_position = pose

            x, y, z = self.xyz
            x += math.cos(self.pitch) * math.cos(self.yaw) * dist
            y += math.cos(self.pitch) * math.sin(self.yaw) * dist
            z += math.sin(self.pitch) * dist
            self.bus.publish('pose2d', [round(x*1000), round(y*1000),
                                        round(math.degrees(self.yaw)*100)])
            self.xyz = x, y, z
        elif channel == 'rotation':
            self.yaw, self.pitch, self.roll = [math.radians(x/100) for x in self.rotation]
        return channel

# vim: expandtab sw=4 ts=4



"""
  Node - processing unit
"""
from threading import Thread

from osgar.bus import BusShutdownException


class Node(Thread):
    """
       Parent classs for all processing "nodes".
    """
    def __init__(self, config, bus):
        super().__init__()
        self.setDaemon(True)
        self.bus = bus
        self.time = None

    def publish(self, channel, data):
        return self.bus.publish(channel, data)

    def listen(self):
        return self.bus.listen()

    def sleep(self, secs):
        self.bus.sleep(secs)

    def is_bus_alive(self):
        return self.bus.is_alive()

    def update(self):
        timestamp, channel, data = self.bus.listen()
        self.time = timestamp
        setattr(self, channel, data)
        return channel

    def run(self):
        try:
            while True:
                self.update()
        except BusShutdownException:
            pass

    def request_stop(self):
        self.bus.shutdown()

# vim: expandtab sw=4 ts=4



from osgar.logger import LogReader, lookup_stream_id
from osgar.lib.serialize import deserialize


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description='Parse logged detected_gas and extract "Gas" artifact')
    parser.add_argument('logfile', help='recorded log file')
    args = parser.parse_args()

    only_stream = lookup_stream_id(args.logfile, 'gas_detector.co2')
    best = None
    with LogReader(args.logfile, only_stream_id=only_stream) as log:
        for timestamp, stream_id, data in log:
            co2 = deserialize(data)
            if best is None or co2 > best[1]:
                print(timestamp, co2)
                best = timestamp, co2
    print('--------------------')
    if best is not None:
        print('best:', best[0], best[1])
    else:
        print("No Gas, sorry")

# vim: expandtab sw=4 ts=4

